/**********************************************************************
 * 
 *  File: 
 * 
 *  Created:
 *
 *  Version: $Id: CgProgram.hpp,v 1.10 2005/09/26 15:07:00 jonhjelm Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#ifndef _CG_PROGRAM_HPP_
#define _CG_PROGRAM_HPP_
#include "shallows/config.h"

#if defined(HAVE_CG) && defined(HAVE_CGGL)

/* To avoid polluting the global namespace with the OpenGL and CG types, is is
 * very important not to include internal/CgProgramImpl.hpp AND
 * internal/CgShader.hpp in this header. They both are declared by forward 
 * declaration and included only in the .cpp file.
 */

#include "shallows/Program.hpp"
#include "shallows/internal/Shader.hpp"
#include "boost/shared_ptr.hpp"

#include <string>
#include <vector>
#include <memory>

namespace shallows {

namespace internal {
	class CgShader;
	class CgProgramImpl; // To avoid exposing Cg-types into the global scope.
}

/** \brief Class for handling GPU programs written in Cg,
 *  where each program consists of a fragment and/or vertex program.
 *  This is the most central class in the shallows library and 
 *  the intention is that parameters are set to the program, 
 *  and then the program is either CgProgram::activate or CgProgram::run 
 *  is called. 
 *
 *  A Program object can read the program either from file,
 *  of from a std::string. The file format is 
 * \code
 * [Vertex shader]
 * 
 * vertex shader source...
 * 
 * [Fragment shader]
 * 
 * fragment shader source...
 * \endcode
 *
 *  Most parameters are forwarded directly to Cg, while 
 *  parameters concerning framebuffers and rendertargets are stored in the 
 *  CgProgram object.
 *
 */

class  DllExport CgProgram : public Program
{
  using Program::read;		       // inject name
public:

   /** Creates an empty cg program */
  CgProgram();

   /** Destroys the cg program */
  virtual ~CgProgram();

  virtual void read(std::string programText, std::string defines);
  
  /** Set the texture to associate with the sampler named name
   */
  virtual void setInputTexture(const char *name,
    boost::shared_ptr<Texture> texture);

   virtual void activate();

   virtual void run();
                 

  /** Returns the compiled version of the program.
   */
  virtual std::string getCompiled();
  virtual void setParam1f(const char *name, const float param);
  virtual void setParam1i(const char *name, const int param);
  virtual void setParam1fv(const char *name, const int count, const float *params);
  virtual void setParam1iv(const char *name, const int count, const int *params);
  virtual void setParam2f(const char *name, const float param0, const float param1);
  virtual void setParam2i(const char *name, const int param0, const int param1);
  virtual void setParam2fv(const char *name, const int count, const float *params);
  virtual void setParam2iv(const char *name, const int count, const int *params);
  virtual void setParam3f(const char *name, const float param0, const float param1, const float param2);
  virtual void setParam3i(const char *name, const int param0, const int param1, const int param2);
  virtual void setParam3fv(const char *name, const int count, const float *params);
  virtual void setParam3iv(const char *name, const int count, const int *params);
  virtual void setParam4f(const char *name, const float param0, const float param1, const float param2, const float param3);
  virtual void setParam4i(const char *name, const int param0, const int param1, const int param2, const int param3);
  virtual void setParam4fv(const char *name, const int count, const float *params);
  virtual void setParam4iv(const char *name, const int count, const int *params);
  virtual void setParamMatrix4fv(const char *name, const float*params);

  virtual void setAttrib1f(const char *name, const float attrib);
  virtual void setAttrib2f(const char *name, const float attrib0, const float attrib1);
  virtual void setAttrib2fv(const char *name, const float *attribs);
  virtual void setAttrib3f(const char *name, const float attrib0, const float attrib1, const float attrib2);
  virtual void setAttrib3fv(const char *name, const float *attribs);
  virtual void setAttrib4f(const char *name, const float attrib0, const float attrib1, const float attrib2, const float attrib3);
  virtual void setAttrib4fv(const char *name, const float *attribs);
  virtual void setAttribPointer(const char *name, int size, int type, unsigned int stride, 
                                const void *pointer, bool normalized = false);
  virtual void enableClientState(const char *name);
  virtual void disableClientState(const char *name);

protected:
		
    void buildUniformsMap( std::string shaderSource );
	void handleIntegerParameter() const;
private:
	// To allow pImpl_ to call addToLog().
	friend class internal::CgProgramImpl; 
	boost::shared_ptr<internal::CgProgramImpl> pImpl_;
};

} //namespace shallows 
#endif // HAVE_CG
#endif // _CG_PROGRAM_HPP_
 
