/**********************************************************************
 * 
 *	File: GLProgram.hpp
 * 
 * 	Created: 20050621
 *
 *  Version: $Id: GLProgram.hpp,v 1.22 2006/02/03 09:14:52 josse77 Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *			 Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *			 Johan Simon Seland <johans@ifi.uio.no>
 *
 * 	This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 * 	This library is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU General Public License
 * 	("GPL") version 2 as published by the Free Software Foundation.
 * 	See the file LICENSE.GPL at the root directory of this source
 * 	distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#ifndef _GLPROGRAM_HPP_
#define _GLPROGRAM_HPP_

#include "shallows/config.h"
#include "shallows/Program.hpp"
#include "shallows/internal/Shader.hpp"
#include "shallows/internal/GLShader.hpp"
#include "shallows/shallowsexcept.hpp"

#include <memory>
#include <stdexcept>

#pragma warning( disable : 4290 )

namespace shallows {

namespace internal {
	//class GLShader;
}

/** \brief Class for handling GPU programs written in OpenGL 
    Shading Language,
 *  where each program consists of a fragment and/or vertex program.
 *  This is the most central class in the shallows library and 
 *  the intention is that parameters are set to the program, 
 *  and then the program is either GLProgram::activate or GLProgram::run 
 *  is called. 
 *
 *  A Program object can read the program either from file,
 *  of from a std::string. The file format is 
 * \code
 * [Vertex shader]
 * 
 * vertex shader source...
 * 
 * [Fragment shader]
 * 
 * fragment shader source...
 * \endcode
 *
 *  Most parameters are forwarded directly to OpenGL, while 
 *  parameters concerning textures, framebuffers and rendertargets 
 *  are stored in the GLProgram object.
 */
class  DllExport GLProgram : public Program {
public:
	GLProgram();
	virtual ~GLProgram();
	
	virtual void read( std::string programText, std::string defines);
	
	/** Set a texture, the texture unit is decided by shallows
	 * \param name The name of the texture specified in the shader.
	 * \param texture Points to the texture.
	*/
	virtual void setInputTexture(const char *name,
				 boost::shared_ptr<Texture> texture);

	/** Set a texture
	 * \param name The name of the texture specified in the shader.
	 * \param texture Points to the texture.
	 * \param texUnit The texture unit where the texture should be
	 * bound.
	*/
	virtual void setInputTexture(const char *name,
				 boost::shared_ptr<Texture> texture,
				     unsigned int texUnit);

  	virtual void activate();
        virtual void run();
  	
	virtual void setParam1f(const char *name, const float param);
  virtual void setParam1i(const char *name, const int param);
  virtual void setParam1fv(const char *name, const int count, const float *params);
  virtual void setParam1iv(const char *name, const int count, const int *params);
  virtual void setParam2f(const char *name, const float param0, const float param1);
  virtual void setParam2i(const char *name, const int param0, const int param1);
  virtual void setParam2fv(const char *name, const int count, const float *params);
  virtual void setParam2iv(const char *name, const int count, const int *params);
  virtual void setParam3f(const char *name, const float param0, const float param1, const float param2);
  virtual void setParam3i(const char *name, const int param0, const int param1, const int param2);
  virtual void setParam3fv(const char *name, const int count, const float *params);
  virtual void setParam3iv(const char *name, const int count, const int *params);
  virtual void setParam4f(const char *name, const float param0, const float param1, const float param2, const float param3);
  virtual void setParam4i(const char *name, const int param0, const int param1, const int param2, const int param3);
  virtual void setParam4fv(const char *name, const int count, const float *params);
  virtual void setParam4iv(const char *name, const int count, const int *params);
  virtual void setParamMatrix4fv(const char *name, const float*params);
  
  virtual void setAttrib1f(const char *name, const float attrib);
  virtual void setAttrib2f(const char *name, const float attrib0, const float attrib1);
  virtual void setAttrib2fv(const char *name, const float *attribs);
  virtual void setAttrib3f(const char *name, const float attrib0, const float attrib1, const float attrib2);
  virtual void setAttrib3fv(const char *name, const float *attribs);
  virtual void setAttrib4f(const char *name, const float attrib0, const float attrib1, const float attrib2, const float attrib3);
  virtual void setAttrib4fv(const char *name, const float *attribs);
  virtual void setAttribPointer(const char *name, int size, int type, unsigned int stride, 
                              const void *pointer, bool normalized = false);
	virtual void enableClientState(const char *name);
	virtual void disableClientState(const char *name);
  	
  /** Return the underlying OpenGL program handle. */
  unsigned int getProgramHandle();
protected:
	unsigned int programHandle_;
	
  typedef std::auto_ptr<internal::GLShader> pGLShader;
	pGLShader vertexShader_;
	pGLShader fragmentShader_;
	
	/** Init a shader. Will throw compile_error if the compile fails.
	 */
	pGLShader initShader( pGLShader shader, internal::Shader::ShaderType type, 
		std::string source );
	
    void buildUniformsMap( std::string shaderSource );
    
	/** Validate that a texture parameter, and that the texture is 
	 * of the same type as expected by the program.
	 * \param name The name of the texture specified in the shader.
	 * \param texture Points to the texture.
	 * \return The location of the uniform
	 */
	int validate_texture( const char* name, 
			      boost::shared_ptr<Texture> texture ) 
	   const throw (std::invalid_argument );

	/** Validate that a uniform name is legal, and does not specify the name
	 *  of a texture.
	 * \return The location of the uniform
	 */
	int validate_uniform( const char* name ) const throw (std::invalid_argument, shallows::shallows_runtime_error );
	
	/** Validate that a vertex attribute name is valid for the current shader.
	 * \return The location of the vertex attribute.
	 */
	 int validate_attrib( const char* name ) const throw (std::invalid_argument);
};
	
} // namespace shallows

#endif //_GLPROGRAM_HPP_
