/**********************************************************************
 * 
 *  File: 
 * 
 *  Created:
 *
 *  Version: $Id: OffScreenBuffer.hpp,v 1.19 2005/12/06 12:48:05 jonhjelm Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#ifndef _SHALLOWS_OFFSCREENBUFFER_HPP
#define _SHALLOWS_OFFSCREENBUFFER_HPP
#include "shallows/config.h"

#include "shallows/FrameBuffer.hpp"
#include <vector>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
namespace shallows {

class Texture2D;	
class RenderTarget;
class OffScreenRenderTarget;
class RenderTexture2D;
class RenderBuffer;

#pragma warning( disable: 4251 )

/** \brief Implements off screen framebuffer using the
 *  framebuffer object extension. 
 *   
 *  Even though FBO's are allowed to change internal format as well
 *  as the dimension of the framebuffer, this class does not.
 *  This is because such changes will in most cases be time
 *  consuming. 
 *  
 *  All functions (including constructor) may set this framebuffer
 *  as current framebuffer.
 *
*/
class  DllExport OffScreenBuffer : public FrameBuffer
{
   friend unsigned int attach(OffScreenBuffer *off_buff, 
		      boost::shared_ptr<shallows::OffScreenRenderTarget> &rt, 
		      int colAttachable);
   friend void attach(OffScreenBuffer *off_buff, 
		     boost::shared_ptr<shallows::OffScreenRenderTarget> *rt, 
		     unsigned int *enums,
		     unsigned int numBuffs);


public:
   /** Creates a floating point framebuffer with four components.
    * \param width The width of the framebuffer.
    * \param height The height of the framebuffer.
    */
   OffScreenBuffer(unsigned int width, unsigned int height);

   /** Creates a framebuffer with user defined format and dimension.
    * \param internalFormat The internalformat used for the 
    * shallows::OffScreenRenderTarget.
    * \param width The width of the framebuffer.
    * \param height The height of the framebuffer.
    */
   OffScreenBuffer(int internalFormat, unsigned int width, 
		   unsigned int height);

   /** Creates a framebuffer with the same dimension and
    *  internal format as the given texture.
    * \param t The texture that specifies the format.
    */
   OffScreenBuffer(Texture2D &t);

    	

   /** Destroys the framebuffer, note that if it is bound as the
    *  current framebuffer it will remain so and therefore not 
    *  really destructed until another framebuffer is activated
    */
   virtual ~OffScreenBuffer();

   /** Creates a shallows::RenderBuffer with the same dimension
    *  and internal format as the framebuffer.
    */
   virtual boost::shared_ptr<RenderBuffer>  createRenderBuffer();

   /** Creates a renderable texture with the same dimension
    *  and internal format as the framebuffer. The texture returned
    *  is of type GL_TEXTURE_2D and texture lookups should be done
    *  by using texture coordinates in the range [0..1]x[0..1].
    */
   virtual boost::shared_ptr<RenderTexture2D>  createRenderTexture2D();

   /** Creates a renderable texture with the same dimension
    *  and internal format as the framebuffer. The texture returned
    *  is of type GL_TEXTURE_RECTANGLE_ARB and texture lookups should be done
    *  by using texture coordinates in the range [0..width]x[0..height].
    */
   virtual boost::shared_ptr<RenderTexture2D> 
     createRenderTextureRectangle();

   /** Creates a depth buffer with the same dimension as the framebuffer. 
    */
   virtual boost::shared_ptr<RenderBuffer>  createDepthBuffer();

   /** Creates a depth buffer which can be used as a texture, 
    * with the same dimension as the framebuffer. 
    * \param rectangular If true the texture will be of type 
    * GL_TEXTURE_RECLANGLE, otherwise it will be of type GL_TEXTURE_2D.
    */
   virtual boost::shared_ptr<RenderTexture2D>  createDepthTexture(bool rectangular=false);

   /** Attaches the rendertarget to the framebuffer. This buffer is also
    *  set as current read and draw buffer. The OpenGL enum specifying the 
    *  attachmeny point (to be used by glDrawBuffer or glReadBuffer) is 
    *  returned.
    *  /param rt The rendertarget to be attached
    *  /param colAttachable The color attachment point  to attach the 
    *  rendertarget to. If it is not specified the rendertarget is attached
    *  to the same attachment point as last time it was attached to this
    *  frabebuffer.
    */
   virtual unsigned int attachRenderTarget(
      boost::shared_ptr<shallows::OffScreenRenderTarget> rt, 
      int colAttachable=-1);

   /** Attaches the rendertargets to the framebuffer. OpenGL is instructed
    *  to use theese targets as output of rendering, by a call to
    *  glDrawBuffers. As a side effect of this function the read buffer 
    *  is set to GL_NONE.
    *  /param rt The rendertargets to be attached
    *  /param numBuffs The number og rendertargets.
    */
   void attachRenderTarget(
      boost::shared_ptr<shallows::OffScreenRenderTarget> *rt, 
      unsigned int numBuffs);

   /** Attaches the rendertarget as a depth buffer. 
    *  /param r The rendertarget to be used as depth buffer, it must have
    *  the same dimension as the framebuffer and the internal format must
    *  be of a type alowed as depth attachment */
   virtual void attachDepthBuffer(RenderTarget &r);

   /** Tests if a rendertarget is compatible with this 
    *  framebuffer, throws shallows_invalid_argument if the 
    *  rendertarget is incompatible.
    *  /param r The rendertarget to validate.
    */
   virtual void validateRenderTarget(const shallows::RenderTarget &r) const;

   /** Sets this offscreen buffer as the current framebuffer.
    */
   virtual void activate();

   /** Checks if the framebuffer is ready to be drawn into and
       throws frame_buffer_incomplete otherwise.
    */
   virtual void validateCompleteness() const;

   /** Returns the width and height of the framebuffer */ 
   virtual void getSize(int &width, int &height) const;

   /** Detaches the rendertarget attached to the given attachement
    *  point. This function is created primarily for internal use.
  */ 
   void detachRenderTarget(int att_point);

protected:
   void setRenderTargetAttached(
      boost::shared_ptr<shallows::OffScreenRenderTarget> &rt, 
      int attachment_point);

   /** Validates whether a render target is attachable to the framebuffer.
    *  Otherwise std::invalid_argument is thrown. 
    * \param r The render target to validate.
    */
   virtual void validateRenderTarget(const OffScreenRenderTarget &r) const;

   virtual std::vector<boost::weak_ptr<OffScreenRenderTarget> > &getTargets();
  private:
   std::vector<boost::weak_ptr<OffScreenRenderTarget> > targets_;
   void init(int internalFormat, unsigned int width, 
	     unsigned int height);
   void setInternalFormat(Texture2D &t);
   void operator=(const OffScreenBuffer &b);
   int height_;
   int width_;
   unsigned int internalFormat_;
   unsigned int depthFormat_;
   unsigned int id_;
   std::vector<unsigned int> drawBuffers_;
};

} // namespace shallows

#endif //  _SHALLOWS_OFFSCREENBUFFER_HPP

