/**********************************************************************
 * 
 *	File: Program.hpp
 * 
 * 	Created: 20050621
 *
 *  Version: $Id: Program.hpp,v 1.23 2006/02/03 09:15:00 josse77 Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *			 Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *			 Johan Simon Seland <johans@ifi.uio.no>
 *
 * 	This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 * 	This library is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU General Public License
 * 	("GPL") version 2 as published by the Free Software Foundation.
 * 	See the file LICENSE.GPL at the root directory of this source
 * 	distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/
 
#ifndef _SHALLOWS_PROGRAM_HPP_
#define _SHALLOWS_PROGRAM_HPP_

#include "shallows/config.h"
#include <string>
#include <vector>
#include <map>
#include <boost/shared_ptr.hpp>
namespace shallows {

class Texture2D;
class Texture;
class FrameBuffer;
class RenderTarget;
class OffScreenRenderTarget;
struct TextureInfo
{
   boost::shared_ptr<Texture> texture;   
   std::string name;
};

struct RenderTargetInfo
{
   boost::shared_ptr<RenderTarget> target;   
   int colorAttachment;
   RenderTargetInfo() : colorAttachment(-1) { ; }
};

/** \brief Abstract base class for handling GPU programs,
 *  where each program consists of a fragment and/or vertex program.
 *  This is the most central class in the shallows library and 
 *  the intention is that parameters are set to the program, 
 *  and then the program is either Program::activate or Program::run 
 *  is called. 
 *
 *  A Program object can read the program either from file,
 *  of from a std::string. The file format is 
 * \code
 * [Vertex shader]
 * 
 * vertex shader source...
 * 
 * [Fragment shader]
 * 
 * fragment shader source...
 * \endcode
 *
 *  Most parameters are forwarded directly to the graphics system
 *  (OpenGL or Cg), while some parameters are stored in the 
 *  Program object. For details on which parameters are forwarded
 *  see the documentation of the inherited class.
 *
 *  For applications that does not use off screen rendering
 *  the program can be used without providing frame buffer.
 */
#pragma warning( disable: 4251 )

class  DllExport Program {
public:
	/** Creates an empty program */
	Program();

	/** Destroys the program */
	virtual ~Program();
	
	/** Reads shaders from file. Calls read() when finished reading from file.
 	\param filename   Filename to file containing shaders. 
	\param defines    A string that will be prepended to the shader before compilation.
 	\throws shallows::file_not_found if filename can not be opened for reading.
 	\throws shallows::compile_error if the shader can not be compiled.
 	*/
  	void readFile( const char *filename, const char *defines=0 );

  	/** Creates program from shaders definded the programText. 
        If one of the shaders fails to compile the exception compile_error
	will be thrown.
  	\param programText String containing shaders.
	\param defines    A string that will be prepended to the shader before compilation.
  	\return bool      Everything went OK. 
	*/
  	virtual void read(std::string programText, std::string defines) = 0;
  	
    /** Create program from shaders.
	\param programText String containing shaders.
    */
  	void read(std::string programText);
  
  	/** Return shader info log. */
  	std::string getLog() const;
  	
  	/** Clear shader info log. */ 
  	void clearLog();


  	/** Activates the program, and ensures that all parameters that
	 *  are stored in the object are forwarded to the rendering
	 *  system. \b Note: this function does \b not change the viewport
	 *  therefore the calling application must ensure that the 
	 *  viewport is set according to the framebuffer in use.
	 */ 
  	virtual void activate() = 0;
 

  	/** Runs the program by rendering the GPGPU quad. 
	 *
  	 *  Before rendering the quad the program is activated and, and
	 *  all parameters that are stored in the object are forwarded 
	 *  to the rendering system. In contrast to Program::activate
	 *  the viewport is set to cover the entire framebuffer 
	 *  if an /b OffScreenBuffer is attached as framebuffer.
	 */ 
        virtual void run() = 0;
   
  	/** Renders the standard GPGPU quad, covering the viewport. 
	 *  The texture coordinates are set to (0, 0) in the lower
	 *  left corner and (width, height) in the upper right corner
	 *  unless otherwise specified by calling 
	 *  Program::setTexCoords or Program::useNormalizedTexCoords.
  	*/
  	virtual void renderQuad();

  	/** Returns the texture coordinates to be used when the GPGPU
	 *  quad is rendered. 
  	*/
  	virtual void getTexCoords(float &min_u, float &min_v, 
				  float &max_u, float &max_v) const;

  	/** Specifies the texture coordinates to be used when the GPGPU
	 *  quad is rendered. 
  	*/
  	virtual void setTexCoords(float min_u, float min_v, 
				  float max_u, float max_v);

  	/** Sets the texture coordinates to (0, 0) in the lower
	 *  left corner and (framebuffer width, framebuffer height) 
	 *  in the upper right corner (default). 
  	*/
  	virtual void useViewportTexCoords();

  	/** Sets the texture coordinates to (0, 0) in the lower
	 *  left corner and (1, 1) 
	 *  in the upper right corner. 
  	*/
  	virtual void useNormalizedTexCoords();

  	/** Fetch the Vertex Shader source from a string. 
  	 */
  	static std::string getVertexShaderSource( std::string programText, 
  		std::string defines = "" );
  	/** Fetch the Fragment Shader source from a string.
  	 */
  	static std::string getFragmentShaderSource( std::string programText,
  		std::string defines = "" );
  	

  	typedef std::map<std::string, std::string> NameTypeMap;
  	

  	/** uniforms allow for examining the uniform variables of a shader.
  	 */
  	const NameTypeMap& uniforms;

   /** Set the framebuffer which the program writes its output 
       \param fb The framebuffer which will get output.
    */
   virtual void setFrameBuffer(boost::shared_ptr<FrameBuffer> fb);

   /** Set a render target that is to be rendered to when the program
       is ran. Output colors are refered to by number instead of
       a logical name in OpenGL.
       \param colorNum specifies the color number (must match a color that
       is rendered to by the fragment program.)
       \param target specifies the render target, usually a renderTexture2D
    */
   void setOutputTarget( unsigned int colorNum, 
			 boost::shared_ptr<RenderTarget> target );
   
   /** Set the texture to be used as input by the program.
       \param name The logical name of the parameter defined 
       by the shader source(s)
       \param texture Pointer to the texture to be sampled.
    */
   virtual void setInputTexture(const char *name,
			  boost::shared_ptr<Texture> texture) = 0;
  	
  	/** Specifies the float value of a uniform variable.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param param      The value of the parameter. */
  virtual void setParam1f(const char *name, const float param) = 0;

  /** Specifies the integer value of a uniform variable.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param param      The value of the parameter. */
  virtual void setParam1i(const char *name, const int param) = 0;

  /** Specifies the float value of a uniform variable.
  \param name Points to a null-terminated string containing the name of the parameter.
  \param count      The number of elements that are to be modified, this should be 1 if the targeted parameter is not an array.
  \param params     Pointer to an array of count values that will be used to update the parameter. */
  virtual void setParam1fv(const char *name, const int count, const float *params) = 0;

  /** Specifies the integer value of a uniform variable.
  \param name Points to a null-terminated string containing the name of the parameter.
  \param count      The number of elements that are to be modified, this should be 1 if the targeted parameter is not an array.
  \param params     Pointer to an array of count values that will be used to update the parameter. */
  virtual void setParam1iv(const char *name, const int count, const int *params) = 0;

  /** Specifies the float values of a uniform variable.
  \param name Points to a null-terminated string containing the name of the parameter.
  \param param0     First component value.
  \param param1     Second component value.  */
  virtual void setParam2f(const char *name, const float param0, const float param1) = 0;

  /** Specifies the integer values of a uniform variable.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param param0     First component value.
  \param param1     Second component value.  */
  virtual void setParam2i(const char *name, const int param0, const int param1) = 0;

  /** Specifies the float values of a uniform variable.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param count      The number of elements that are to be modified, this should be 1 if the targeted parameter is not an array.
  \param params     Pointer to an array of count values that will be used to update the parameter. */
  virtual void setParam2fv(const char *name, const int count, const float *params) = 0;

  /** Specifies the integer values of a uniform variable.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param count      The number of elements that are to be modified, this should be 1 if the targeted parameter is not an array.
  \param params     Pointer to an array of count values that will be used to update the parameter. */
  virtual void setParam2iv(const char *name, const int count, const int *params) = 0;
  
  /** Specifies the float values of a uniform variable.
  \param name Points to a null-terminated string containing the name of the parameter.
  \param param0     First component value.
  \param param1     Second component value. 
  \param param2     Third component value.  */
  virtual void setParam3f(const char *name, const float param0, const float param1, const float param2) = 0;

  /** Specifies the integer values of a uniform variable.
  \param name Points to a null-terminated string containing the name of the parameter.
  \param param0     First component value.
  \param param1     Second component value. 
  \param param2     Third component value.  */
  virtual void setParam3i(const char *name, const int param0, const int param1, const int param2) = 0;
  
  /** Specifies the float values of a uniform variable.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param count      The number of elements that are to be modified, this should be 1 if the targeted parameter is not an array.
  \param params     Pointer to an array of count values that will be used to update the parameter. */
  virtual void setParam3fv(const char *name, const int count, const float *params) = 0;
  
  /** Specifies the integer values of a uniform variable.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param count      The number of elements that are to be modified, this should be 1 if the targeted parameter is not an array.
  \param params     Pointer to an array of count values that will be used to update the parameter. */
  virtual void setParam3iv(const char *name, const int count, const int *params) = 0;
  
  /** Specifies the float values of a uniform variable.
  \param name Points to a null-terminated string containing the name of the parameter.
  \param param0     First component value.
  \param param1     Second component value. 
  \param param2     Third component value. 
  \param param3     Fourth component value. */
  virtual void setParam4f(const char *name, const float param0, const float param1, const float param2, const float param3) = 0;
  
  /** Specifies the integer values of a uniform variable.
  \param name Points to a null-terminated string containing the name of the parameter.
  \param param0     First component value.
  \param param1     Second component value. 
  \param param2     Third component value. 
  \param param3     Fourth component value. */
  virtual void setParam4i(const char *name, const int param0, const int param1, const int param2, const int param3) = 0;
  
  /** Specifies the float values of a uniform variable.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param count      The number of elements that are to be modified, this should be 1 if the targeted parameter is not an array.
  \param params     Pointer to an array of count values that will be used to update the parameter. */
  virtual void setParam4fv(const char *name, const int count, const float *params) = 0;
  
  /** Specifies the integer values of a uniform variable.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param count      The number of elements that are to be modified, this should be 1 if the targeted parameter is not an array.
  \param params     Pointer to an array of count values that will be used to update the parameter. */
  virtual void setParam4iv(const char *name, const int count, const int *params) = 0;
  

  /** Specifies the float values of a uniform matrix.
  \param name       Points to a null-terminated string containing the name of the parameter.
  \param params     Pointer to an array of values that will be used to update the matrix. */
  virtual void setParamMatrix4fv(const char *name, const float *params) = 0;
  
  /** Specifies the float value of a vertex attribute.
  \param name       Points to a null-terminated string containing the name of the attribute.
  \param attrib     The value of the attribute. */
  virtual void setAttrib1f(const char *name, const float attrib) = 0;

  /** Specifies the float values of a vertex attribute.
  \param name       Points to a null-terminated string containing the name of the attribute.
  \param attrib0    First component value. 
  \param attrib1    Second component value. */
  virtual void setAttrib2f(const char *name, const float attrib0, const float attrib1) = 0;

  /** Specifies the float values of a vertex attribute.
  \param name       Points to a null-terminated string containing the name of the attribute.
  \param attribs    Pointer to an array of values to be used for the attribute. */
  virtual void setAttrib2fv(const char *name, const float *attribs) = 0;

  /** Specifies the float values of a vertex attribute.
  \param name       Points to a null-terminated string containing the name of the attribute.
  \param attrib0    First component value. 
  \param attrib1    Second component value. 
  \param attrib2    Third component value. */
  virtual void setAttrib3f(const char *name, const float attrib0, const float attrib1, const float attrib2) = 0;

  /** Specifies the float values of a vertex attribute.
  \param name       Points to a null-terminated string containing the name of the attribute.
  \param attribs    Pointer to an array of values to be used for the attribute. */
  virtual void setAttrib3fv(const char *name, const float *attribs) = 0;

  /** Specifies the float values of a vertex attribute.
  \param name       Points to a null-terminated string containing the name of the attribute.
  \param attrib0    First component value. 
  \param attrib1    Second component value. 
  \param attrib2    Third component value. 
  \param attrib3    Fourth component value. */
  virtual void setAttrib4f(const char *name, const float attrib0, const float attrib1, const float attrib2, const float attrib3) = 0;

  /** Specifies the float values of a vertex attribute.
  \param name       Points to a null-terminated string containing the name of the attribute.
  \param attribs    Pointer to an array of values to be used for the attribute. */
  virtual void setAttrib4fv(const char *name, const float *attribs) = 0;


  /** Defines a generic vertex attribute array.
  \param name       Points to a null-terminated string containing the name of the attribute.
  \param size       Specifies the number of values for each element of the generic vertex attribute array. Must be 1, 2, 3 og 4.
  \param type       Specities the data type of each component in the array. Symbolic constants GL_BYTE, GL_UNSIGNED_BYTE, GL_SHORT, 
                    GL_UNSIGED_SHORT, GL_INT, GL_UNSIGNED_INT, GL_FLOAT and GLDOUBLE are accepted.
  \param stride     Specifies the bye offset between consecutive attribute values. If stride is 0, the attribute values are understood 
                    to be tightly packed in the array.
  \param pointer    The pointer to the first component of the first attribute in the array.
  \param normalized Specifies whether fixed-point data values should be normalized (GL_TRUE) or converted directly as fixed-point 
                    values ( GL_FALSE ) when they are accessed. */
  virtual void setAttribPointer(const char *name, int size, int type, 
                                unsigned int stride, const void *pointer, bool normalized = false) = 0;


  /** Enables a generic vertex attribute array.
  \param name       Points to a null-terminated string containing the name of the vertex attribute array. */
  virtual void enableClientState(const char *name) = 0;

  /** Disables a generic vertex attribute array.
  \param name       Points to a null-terminated string containing the name of the vertex attribute array. */
  virtual void disableClientState(const char *name) = 0;
   
protected:
	
   /** Binds all rendertargets to the corresponding color attachable */
  	virtual void activateRenderTargets();

   /** Adds a string to the event log */
	void addToLog(const std::string &s);	
	
	std::string log_;
	
	/** Build a map of the uniforms. */
	virtual void buildUniformsMap( std::string shaderSource ) = 0;
	
	/** uniformsNameToType store the name and type of every unifom in the 
	 * program. 
	 */
	NameTypeMap uniformsNameToType_;

   std::string printUniformsMap() const;
   /** States whether texturecoords should user specified, or follow 
    *  viewport, when renderQuad is used */
   bool useViewportTexCoords_;
   float min_u_, min_v_, max_u_, max_v_;
   std::vector< shallows::TextureInfo > textureInfo_;
   //   std::vector< boost::shared_ptr<shallows::RenderTarget> > renderTargets_;
//   std::vector< shallows::RenderTargetInfo > renderTargetInfo_;
   /** vector indicating which color attachments are in use
    * and which output colors they correspond to. */
//   std::vector< int > colorAttachment_;
   std::vector< boost::shared_ptr<RenderTarget> > renderTarget_;
   boost::shared_ptr< shallows::FrameBuffer > frameBuffer_;
  
};
	
} // namespace shallows
#endif //_SHALLOWS_PROGRAM_HPP_
