/**********************************************************************
 * 
 *  File: file
 * 
 *  Created: Jul 14, 2005
 *
 *  Version: $Id: imageCreationInternal.hpp,v 1.6 2005/09/27 08:56:38 josse77 Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/
#ifndef _IMAGECREATIONINTERNAL_HPP_
#define _IMAGECREATIONINTERNAL_HPP_

#include "shallows/config.h"

#ifdef HAVE_MAGICK
#include <Magick++.h>
#endif

#ifdef HAVE_PNG
#include <png.h>
#include <cstdlib>
#include <vector>
#include <streambuf>
#include <ios>
#include <fstream>
#endif

#include "shallows/utils/Image.hpp"
#include "shallows/shallowsexcept.hpp"
#include "shallows/utils/utilsexcept.hpp"

#include <boost/shared_ptr.hpp>
#include <string>

namespace shallows {
namespace internal {

using shallows::utils::Image;

#ifdef HAVE_MAGICK
/** Map type to an ImageMagick enum.
 * These funky types only exists as a template specializations, 
 * trying to compile with other parameters result in compiler error.
 * 
 * The rationale is that we avoid needing specializations in genIMImage.
 */
 
template<typename T>
struct type2IMenum {
    Magick::StorageType operator()();
};


std::string
channelsToIMString( const unsigned int channels );

template<typename T>
boost::shared_ptr<Image<T> > 
genIMImage( std::string filename, unsigned int channels ) {
    // Map T to an ImageMagick::enum
    Magick::Image image(filename); // Will throw if file could not be loaded.
    image.flip(); // To get sane OGL order.
    
    boost::shared_ptr<Image<T> > res( new Image<T>() );
    
    res->width_ = image.columns();
    res->height_ = image.rows();
    res->channels_ = channels;
    res->data_.resize( image.columns() * image.rows() * channels );
    
    image.write( 0, 0, image.columns(), image.rows(), channelsToIMString( channels ),
        type2IMenum<T>()(), &(res->data_[0]) );

    return res;
}

template<typename T>
void
writeImageToFile( const Image<T>& I, std::string filename ) {
    Magick::Image image( I.width_, I.height_, channelsToIMString( I.channels_ ), 
        type2IMenum<T>()(), &(I.data_[0]) );
    image.flip();
    image.write( filename );
}

#endif // HAVE_MAGICK

#ifdef HAVE_PNG


/** type2PngTransforms tries to set up correct conversions for the correct type
 *  for use with libpng. Only specializations exist.
 */
template<typename T>
int  
type2PngTransforms( unsigned int channels ) {
    throw shallows::utils::not_supported( "" );
}

template<>
int
type2PngTransforms<char>( unsigned int channels );

template<>
int
type2PngTransforms<short>( unsigned int channels );


template<>
int
type2PngTransforms<int>( unsigned int channels );

/** pnglib packs everything into bytes. So even if we read 16-bit images, the 
 *  result will be RRGGBB RRGGBB in bytes. This function tries to be smart
 *  and bitshifts bytes in T. 
 *  \warning This function only works as long as a png_byte is 8-bit.
 */
template<typename T>
void
datainserter( std::vector<T>& v, png_byte row[], 
    unsigned int channels, unsigned int depth, std::size_t width ) {
    
    const int toaccum = depth/8;

     if ( sizeof(T) < sizeof(png_byte)*toaccum ) 
	 throw std::runtime_error( "Type of container it not large enough to hold all data." );
    
    for ( std::size_t i = 0; i < width*channels*toaccum; ) {
        T value = 0;
        for ( int j = toaccum-1; j >= 0; --j ) 
            value += static_cast<T> (row[i++] << j*8); // Assume png_byte is 8-bit.
	
        v.push_back( value );         
    }
}

/** Read a png file into an Image object. This function support 1, 2, 4, 8 and 16 bit
 *  png files. Reading into a container of a type with higher bit order than the file will
 *  _NOT_ result in upscaling the data. 
 *  ie. 1, 2 and 4 will not be scaled into the full 8-bit range when reading into a char.
 *  Reading a 16-bit image int an int-containter will not scale the value either.
 *  
 *  However reading a 16-bit png file into an 8-bit container  will trigger scaling into 
 *  that range.  
 * 
 *  \warning This function calls type2PngTransforms, so it will only work as long as 
 *  png_byte is 8-bit.
 */
template<typename T>
boost::shared_ptr<Image<T> > 
genPngImage( std::string filename, unsigned int channels ) {
    using namespace std;
    
    boost::shared_ptr<Image<T> > res( new Image<T>() );
    
    // libpng's basic read-callback expects a FILE pointer.
    // We do this c-style for now. (Though I don't like it.)
    FILE *fp = fopen( filename.c_str(), "rb" );
    if (!fp) 
        throw shallows::file_not_found( filename );
    
    vector<png_byte> header(8);
    fread( &header[0], 1, header.size(), fp );
    
    // Check for png file format
    if (png_sig_cmp( &header[0], 0, header.size() ) ) {
        fclose( fp );
        throw shallows::utils::not_supported( filename + " is not of valid file format for pnglib. " );
    }
    
    png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
    if (!png_ptr) {
        fclose( fp );
        throw std::runtime_error( "Trouble using pnglib" );
    }
        
    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr ) {
        fclose( fp );
        throw std::runtime_error( "Trouble using pnglib" );
    }
   
   png_infop end_info = png_create_info_struct(png_ptr);
   if (!end_info) {
        fclose( fp );
        png_destroy_read_struct(&png_ptr, &info_ptr,
            (png_infopp)NULL);
        throw std::runtime_error( "Trouble using pnglib" );
    } 
        
    png_init_io( png_ptr, fp );
    png_set_sig_bytes(png_ptr, header.size() );    
    
    png_read_png( png_ptr, info_ptr, type2PngTransforms<T>( channels ), 0 );
    fclose( fp );
    
    res->width_ = info_ptr->width;
    res->height_ = info_ptr->height;
    res->channels_ = info_ptr->channels;
    
    png_bytep *row_pointers;
    row_pointers = png_get_rows(png_ptr, info_ptr);
    
    for ( int i = info_ptr->height - 1; i >= 0 ; --i )
        datainserter( res->data_, row_pointers[i], info_ptr->channels,
            info_ptr->bit_depth, info_ptr->width );
    return res;
}

#endif // HAVE_CG

} // namespace internal
} // namespace shallows
#endif //_IMAGECREATIONINTERNAL_HPP_
