/**********************************************************************
 * 
 *  File: 
 * 
 *  Created:
 *
 *  Version: $Id: shallows.hpp,v 1.18 2005/10/20 15:04:32 josse77 Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/
#ifndef _SHALLOWS_HPP
#define _SHALLOWS_HPP

#pragma warning( disable : 4290 )

#include "shallows/config.h"

#include <stdexcept>
#include <string>

/** \mainpage
    shallows is a library to ease the development of programs using
    shading languages and OpenGL. Both the OpenGL shading language (GLSlang)
    and Nvidias Cg are supported. More specifically, the library is aimed at 
    easing the development of GPGPU (General Purpose computation on Graphical
    Processing Units) programs.
    
    The rationale of shallows is to expose functinality which maps closely to an
    experienced OpenGL developers mental image of the pipeline.
    
    Documented examples are in the following pages:
    - <a href="simpleHeat_8cpp-example.html">simpleHeat.cpp</a> The basic Hello GPGPU application
    - <a href="simpleMRT_8cpp-example.html">simpleMRT.cpp</a> Demonstrates the use of multiple render targets
    - <a href="simpleWave_8cpp-example.html">simpleWave.cpp</a> A wave equation with fancy rendering

 *  Copyright (C) 2005 by SINTEF.  All rights reserved.\n
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.

*/

#include "shallows/Program.hpp"
#include "shallows/CgProgram.hpp"
#include "shallows/GLProgram.hpp"
#include "shallows/FrameBuffer.hpp"
#include "shallows/OffScreenBuffer.hpp"
#include "shallows/OffScreenRenderTarget.hpp"
#include "shallows/OnScreenBuffer.hpp"
#include "shallows/OnScreenRenderTarget.hpp"
#include "shallows/RenderTarget.hpp"
#include "shallows/RenderTexture2D.hpp"
#include "shallows/RenderBuffer.hpp"
#include "shallows/shallowsexcept.hpp"
#include "shallows/Texture.hpp"
#include "shallows/Texture1D.hpp"
#include "shallows/Texture2D.hpp"
#include "shallows/Texture3D.hpp"

/** \example simpleWave.cpp */
/** \example simpleHeat.cpp */
/** \example simpleMRT.cpp */

/** \namespace shallows
 *  \brief shallows is the main namespace for the library. 
 * 
 *  Members of this namespace should only reference libraries in the requirements of
 *  shallows, and the API should be considered stable.
 */
 
 /** \namespace shallows::internal
  *  \brief internal holds implementation details of classes in shallows.
  * 
  *  Members of the internal namespace may change without notice. This is
  *  strictly an implementational issue, and no code should rely on details
  *  of this namespace.
  */
namespace shallows {
	
	/** Activate the OpenGL fixed function pipeline. */
	DllExport
	void
	setFixedFunction();

	/** Initialize the shallows library, including OpenGL extenstions.
	    important for Windows compatibility. */
	DllExport
	void
	init_shallows();
	

	/** Check if OpenGL has flagged an error, and throw and exception if it has. 
	\param filename The current file
	\param line The current line
	*/
	DllExport
	void
	check_GL( std::string filename = "", int line = 0 ) throw ( std::runtime_error );
    
    /** Check if the Cg runtime has flagged an error, and throw and exception if it has.
     \param filename The current file
     \param line The current line
     */
     DllExport
     void
     check_Cg( std::string filename = "", int line = 0 ) throw ( std::runtime_error );


} // namespace shallows

/** CHECK_GL will call shallows::check_GL with the ANSI C 
 * __FILE__ and __LINE__ macros. This should give meaningful error messages.
 */
#define CHECK_GL shallows::check_GL( __FILE__, __LINE__ );

#define CHECK_CG shallows::check_Cg( __FILE__, __LINE__ );

/** CHECK_GL_DEBUG is intended for use as a postcondition for every 
 * function we call that may alter the GL-state. As the last statement before
 * return they should call this macro, which in debug mode will raise an error 
 * by throwing. In release mode this macro does nothing. This is the same
 * behaviour as the class C-assert macro.
 */
#ifdef DEBUG
#define CHECK_GL_DEBUG shallows::check_GL( __FILE__, __LINE__ );
#define CHECK_CG_DEBUG shallows::check_Cg( __FILE__, __LINE__ );
#else
#define CHECK_GL_DEBUG ;
#define CHECK_CG_DEBUG ;
#endif

#endif // _SHALLOWS_HPP
