/**********************************************************************
 * 
 *  File: file
 * 
 *  Created: Jul 13, 2005
 *
 *  Version: $Id: Image.hpp,v 1.11 2005/10/21 09:03:43 josse77 Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/
#ifndef _IMAGE_HPP_
#define _IMAGE_HPP_

#include "shallows/config.h"
#include <vector>
#include <boost/shared_ptr.hpp>
#include <boost/static_assert.hpp>
#include <shallows/Texture2D.hpp>

namespace shallows
{

namespace utils
{

/** \brief Image is a small, lightweight image class.
 * 
 *  Image is a small, lightweight image class.
 *  Image provides litte more than a collection of data and geometry.
 *  The rationale behind this class is that is should be easy to write 
 *  separate image loader classes, and then send them into shallows.
 * 
 *  It is possible to alter data and geometry by the user, but no sanity 
 *  checking is done on these alterations.
 */
 
template<typename T>
class Image
{
public:
    
    /** Base Image constructor. It will copy the data fra T* data, and
     *  create an image with of the given size. 
     */
	Image( unsigned int width, unsigned int height, unsigned int channels,
        const T* data) : 
        width_(width),
        height_(height),
        channels_(channels),
        data_( data, data+(width*height*channels) )
        {};
    
    /** Basically the same as the constructor above, but using vectors instead
     *  of pointer. 
     */
    Image( unsigned int width, unsigned int height, unsigned int channels,
        const std::vector<T>& data ) :
        width_(width),
        height_(height),
        channels_(channels),
        data_(data) 
        {};
    
    /** Create an empty image of the given size. This c'tor will allocate the
     *  given size, but not fill it with anything. 
     */
    Image( unsigned int width, unsigned int height, unsigned int channels ) : 
        width_(width),
        height_(height),
        channels_(channels),
        data_(width*height*channels) 
        {};
        
    /** Create an empty image of size 0. No data is allocated, 
     *  and all fields are initialized to 0. \
     */
    Image() :
        width_(0),
        height_(0),
        channels_(0),
        data_(0)
        {};
        
	virtual ~Image() {};
    
    /** Create a texture of the given image. */
    boost::shared_ptr<shallows::Texture2D> createTexture( 
        unsigned int internalGLformat, bool rectangular = false ) const;
    
    /** Width of image. */
    unsigned int width_;
    /** Height of image. */
    unsigned int height_;
    /** Channels in the image. */
    unsigned int channels_;
    /** The internal data. */
    std::vector<T> data_;
};

/** Map a number of channels to the appropiate GL-enum. */
DllExport 
unsigned int
channelsToGLenum(  const unsigned int channels );

/** Map a type to an OpenGL enum. Only specialized. */

template<typename T>
DllExport
unsigned int
type2GLExternalFormat();

template<>
DllExport
unsigned int
type2GLExternalFormat<double>();

template<>
DllExport
unsigned int
type2GLExternalFormat<float>();

template<>
DllExport
unsigned int
type2GLExternalFormat<int>();

template<>
DllExport
unsigned int
type2GLExternalFormat<unsigned int>();

template<>
DllExport
unsigned int
type2GLExternalFormat<char>();

template<>
DllExport
unsigned int
type2GLExternalFormat<unsigned char>();

template<typename T>
boost::shared_ptr<shallows::Texture2D>
Image<T>::createTexture( 
    unsigned int internalGLformat, bool rectangular ) const {

    unsigned int format = channelsToGLenum( channels_ );
    
    boost::shared_ptr<shallows::Texture2D> tex( 
        new shallows::Texture2D(
            internalGLformat,
            width_, height_, 
            format,
            shallows::utils::type2GLExternalFormat<T>(), 
            &data_[0],
            rectangular ) );
   
   return tex;
}
    
}; // namespace utils

}; // namespace shallows

#endif //_IMAGE_HPP_
