/**********************************************************************
 * 
 *  File: CgProgram.cpp
 * 
 *  Created: 20050623
 *
 *  Version: $Id: CgProgram.cpp,v 1.13 2005/09/20 10:23:09 jonhjelm Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#include "shallows/config.h"

#if defined(HAVE_CG) && defined(HAVE_CGGL)

#include "shallows/shallows.hpp"
#include "shallows/CgProgram.hpp"
#include "shallows/shallowsexcept.hpp"
#include "shallows/internal/CgShader.hpp"
#include "shallows/internal/GLIncludes.hpp"
#include "shallows/internal/CgProgramImpl.hpp"
#include "shallows/FrameBuffer.hpp"
#include <cassert>


namespace shallows {
using namespace std;
using internal::CgShader;
using internal::Shader;

CgProgram::CgProgram() :
	pImpl_( new internal::CgProgramImpl( *this ) )
{
}

CgProgram::~CgProgram()
{
	// vertexShader_ and fragmentShader_ are auto_ptr's so they will
	// be deleted by the auto_ptr's d'tor.
}

void 
CgProgram::read(std::string programText, std::string defines)
{
  	pImpl_->context_ = cgCreateContext();
  	//  cgGLSetManageTextureParameters(context_, true);
  	pImpl_->vertexProfile_ = cgGLGetLatestProfile(CG_GL_VERTEX);
  	pImpl_->fragmentProfile_ = cgGLGetLatestProfile(CG_GL_FRAGMENT);
  	log_.clear();
	
	string vertexShaderSource = Program::getVertexShaderSource( programText, defines );
	string fragmentShaderSource = Program::getFragmentShaderSource( programText, defines );
	
	// These will throw compile error if something is wrong.
	// The semantics here is due to the shaders being auto_ptr's.
	pImpl_->vertexShader_ = pImpl_->initShader( pImpl_->vertexShader_, Shader::VertexShader, pImpl_->context_, pImpl_->vertexProfile_, 
		vertexShaderSource );
	pImpl_->fragmentShader_ = pImpl_->initShader( pImpl_->fragmentShader_, Shader::FragmentShader, pImpl_->context_, pImpl_->fragmentProfile_,
		fragmentShaderSource );
	
    buildUniformsMap ( defines + programText );

	CHECK_GL_DEBUG;
    CHECK_CG_DEBUG;
}

void 
CgProgram::setInputTexture(const char *name,
    boost::shared_ptr<Texture> texture) {
    CGparameter cgp = pImpl_->validateTexture(name, *texture );
    cgGLSetTextureParameter( cgp, texture->getTextureID() );
    
    CHECK_GL_DEBUG;
    CHECK_CG_DEBUG;
    
}
    
void 
CgProgram::activate()
{
   activateRenderTargets();
   CgShader *shader;
   int i;
   for (i=0, shader=(pImpl_->vertexShader_.get()); 
	i<2; shader=pImpl_->fragmentShader_.get(), ++i)
   {
      if (shader==0)
      {
	 cgGLDisableProfile((i==0)?CG_PROFILE_ARBVP1:CG_PROFILE_ARBFP1);
	 continue;
      }
      shader->enable();
   }  

}

void 
CgProgram::run()
{
   activate();
   const OffScreenBuffer *off_buf = 
      dynamic_cast<OffScreenBuffer *>(frameBuffer_.get());
   if (off_buf != 0)
   {
      int width, height;
      glPushAttrib(GL_VIEWPORT_BIT);
      off_buf->getSize(width, height);
      glViewport(0, 0, width, height);
   }      
   renderQuad();
   if (off_buf != 0)
   {
      glPopAttrib();
   }      
   CHECK_GL_DEBUG;
   CHECK_CG_DEBUG;
}
    
std::string CgProgram::getCompiled()
{
  std::string str;
  if ( pImpl_->vertexShader_.get() )
  {
    str += "[Vertex Shader]\n";
    str +=  cgGetProgramString(pImpl_->vertexShader_->getShaderID(), CG_COMPILED_PROGRAM);
    str += "\n";
  }
  if ( pImpl_->fragmentShader_.get())
  {
    str += "[Fragment Shader]\n";
    str += cgGetProgramString(pImpl_->fragmentShader_->getShaderID(), CG_COMPILED_PROGRAM); 
    str += "\n";
 }
 CHECK_CG_DEBUG;
  return str;
}

void CgProgram::setParam1f(const char *name, const float param)
{
  CGparameter cgp = pImpl_->validateUniform( name );
  cgGLSetParameter1f(cgp, param);
  CHECK_CG_DEBUG;
}

void CgProgram::setParam1i(const char *name, const int param)
{
	handleIntegerParameter();
}

void CgProgram::setParam1fv(const char *name, const int count, const float *params)
{
  CGparameter cgp = pImpl_->validateUniform( name );
  if(cgp != 0) 
  {
    if(count == 1)
    {
      cgGLSetParameter1fv(cgp, params);
    }
    else
    {
      cgGLSetParameterArray1f(cgp, 0, count, params);
    }
  }
  CHECK_CG_DEBUG;
}

void CgProgram::setParam1iv(const char *name, const int count, const int *params)
{
  handleIntegerParameter();
}

void CgProgram::setParam2f(const char *name, const float param0, const float param1)
{
  CGparameter cgp = pImpl_->validateUniform( name );
  cgGLSetParameter2f(cgp, param0, param1);
  CHECK_CG_DEBUG;
}

void CgProgram::setParam2i(const char *name, const int param0, const int param1)
{
	handleIntegerParameter();
}

void CgProgram::setParam2fv(const char *name, const int count, const float *params)
{
  CGparameter cgp = pImpl_->validateUniform( name );
  if(cgp != 0) 
  {
    if(count == 1)
    {
      cgGLSetParameter2fv(cgp, params);
    }
    else
    {
      cgGLSetParameterArray2f(cgp, 0, count, params);
    }
  }
  CHECK_CG_DEBUG;
}

void CgProgram::setParam2iv(const char *name, const int count, const int *params)
{
	handleIntegerParameter();
}

void CgProgram::setParam3f(const char *name, const float param0, const float param1, const float param2)
{
  CGparameter cgp = pImpl_->validateUniform( name );
  cgGLSetParameter3f(cgp, param0, param1, param2);
  CHECK_CG_DEBUG;
}

void CgProgram::setParam3i(const char *name, const int param0, const int param1, const int param2)
{
	handleIntegerParameter();
}

void CgProgram::setParam3fv(const char *name, const int count, const float *params)
{
  CGparameter cgp = pImpl_->validateUniform( name );
  if(cgp != 0) 
  {
    if(count == 1)
    {
      cgGLSetParameter3fv(cgp, params);
    }
    else
    {
      cgGLSetParameterArray3f(cgp, 0, count, params);
    }
  }
  CHECK_CG_DEBUG;
}

void CgProgram::setParam3iv(const char *name, const int count, const int *params)
{
	handleIntegerParameter();
}

void CgProgram::setParam4f(const char *name, const float param0, const float param1, const float param2, const float param3)
{
  CGparameter cgp = pImpl_->validateUniform( name );
  cgGLSetParameter4f(cgp, param0, param1, param2, param3);
  CHECK_CG_DEBUG;
}

void CgProgram::setParam4i(const char *name, const int param0, const int param1, const int param2, const int param3)
{
	handleIntegerParameter();
}

void CgProgram::setParam4fv(const char *name, const int count, const float *params)
{
  CGparameter cgp = pImpl_->validateUniform( name );
  if(cgp != 0) 
  {
    if(count == 1)
    {
      cgGLSetParameter4fv(cgp, params);
    }
    else
    {
      cgGLSetParameterArray4f(cgp, 0, count, params);
    }
  }
  CHECK_CG_DEBUG;
}

void CgProgram::setParam4iv(const char *name, const int count, const int *params)
{
	handleIntegerParameter();
}

void CgProgram::setParamMatrix4fv(const char *name, const float *params)
{
	throw not_implemented( __FUNCTION__ );
}

void CgProgram::setAttrib1f(const char *name, const float attrib)
{
  CGparameter cgp = pImpl_->validateAttrib( name );
  cgGLSetParameter1f( cgp, attrib );
  CHECK_CG_DEBUG;
}

void CgProgram::setAttrib2f(const char *name, const float attrib0, const float attrib1)
{
  CGparameter cgp = pImpl_->validateAttrib( name );
  cgGLSetParameter2f( cgp, attrib0, attrib1 );
  CHECK_CG_DEBUG;
}

void CgProgram::setAttrib2fv(const char *name, const float *attribs)
{
  CGparameter cgp = pImpl_->validateAttrib( name );  
  cgGLSetParameter2fv(cgp, attribs);
  CHECK_CG_DEBUG;
}

void CgProgram::setAttrib3f(const char *name, const float attrib0, const float attrib1, const float attrib2)
{
  CGparameter cgp = pImpl_->validateAttrib( name );
  cgGLSetParameter3f( cgp, attrib0, attrib1, attrib2 );
  CHECK_CG_DEBUG;
}

void CgProgram::setAttrib3fv(const char *name, const float *attribs)
{
  CGparameter cgp = pImpl_->validateAttrib( name );
  cgGLSetParameter3fv( cgp, attribs );
  CHECK_CG_DEBUG;
}
void CgProgram::setAttrib4f(const char *name, const float attrib0, const float attrib1, const float attrib2, const float attrib3)
{
  CGparameter cgp = pImpl_->validateAttrib( name );
  cgGLSetParameter4f( cgp, attrib0, attrib1, attrib2, attrib3 );
  CHECK_CG_DEBUG;
}

void CgProgram::setAttrib4fv(const char *name, const float *attribs)
{
    CGparameter cgp = pImpl_->validateAttrib( name );
    cgGLSetParameter4fv( cgp, attribs );
  CHECK_CG_DEBUG;
}

void CgProgram::setAttribPointer(const char *name, int size, int type, 
				 unsigned int stride, 
                                 const void *pointer, bool normalized)
{
  	CGparameter cgp = pImpl_->validateAttrib( name );
  	cgGLSetParameterPointer(cgp, size, type, stride, (GLvoid*)(pointer));
    CHECK_CG_DEBUG;
}

void CgProgram::enableClientState(const char *name)
{
  CGparameter cgp = pImpl_->validateAttrib( name );
  cgGLEnableClientState(cgp);
  CHECK_CG_DEBUG;
}

void CgProgram::disableClientState(const char *name)
{
  CGparameter cgp = pImpl_->validateAttrib( name );
  cgGLDisableClientState(cgp);
  CHECK_CG_DEBUG;
}


void
CgProgram::handleIntegerParameter() const {
	throw std::invalid_argument ( "Unable to set integer parameters in Cg" );	
}

void 
CgProgram::buildUniformsMap( std::string shaderSource ) {
    pImpl_->buildUniformsMap();
}
} //namespace shallows
#endif // HAVE_CG
