/**********************************************************************
 * 
 *	File: GLProgram.cpp
 * 
 * 	Created: 20050621
 *
 *  Version: $Id: GLProgram.cpp,v 1.31 2006/02/03 09:13:51 josse77 Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *			 Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *			 Johan Simon Seland <johans@ifi.uio.no>
 *
 * 	This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 * 	This library is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU General Public License
 * 	("GPL") version 2 as published by the Free Software Foundation.
 * 	See the file LICENSE.GPL at the root directory of this source
 * 	distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#include "shallows/GLProgram.hpp"
#include "shallows/internal/GLShader.hpp"
#include "shallows/shallows.hpp"
#include "shallows/shallowsexcept.hpp"
#include "boost/regex.hpp"
#include "shallows/internal/GLIncludes.hpp"
#include "shallows/Texture.hpp"
#include "shallows/FrameBuffer.hpp"
#include "shallows/OnScreenBuffer.hpp"

#include <iostream>
#include <sstream>
using std::string;
using namespace shallows::internal;

namespace shallows {

GLProgram::GLProgram()
{
	programHandle_ = glCreateProgram();
	
	CHECK_GL_DEBUG;
}

GLProgram::~GLProgram() {
	glDeleteProgram( programHandle_ );
	
	CHECK_GL_DEBUG;
}

void
GLProgram::read( const std::string programText, const std::string defines ) {
	if ( vertexShader_.get() != 0 ) {
		glDetachShader( programHandle_, vertexShader_->getShaderID() );
		vertexShader_.reset( NULL );
	}
	assert( vertexShader_.get() == 0 );
	
	if ( fragmentShader_.get() != 0 ) {
		glDetachShader( programHandle_, fragmentShader_->getShaderID() );
		fragmentShader_.reset( NULL );
	}
	assert( fragmentShader_.get() == 0 );
	
	log_.clear();
	
	buildUniformsMap( defines + programText );		
	
	string vertexShaderSource = Program::getVertexShaderSource( programText, defines );
	string fragmentShaderSource = Program::getFragmentShaderSource( programText, defines );
	
	// These will throw compile error if something is wrong.
	// The semantics here is due to the shaders being auto_ptr's.
	vertexShader_ = initShader( vertexShader_, internal::Shader::VertexShader, vertexShaderSource );
    fragmentShader_ = initShader( fragmentShader_, internal::Shader::FragmentShader, fragmentShaderSource );
		
	char str[128*4096]; // For GL-state reporting
	int result;
		
	glLinkProgram(programHandle_);
	glGetObjectParameterivARB(programHandle_, GL_OBJECT_LINK_STATUS_ARB, &result);
  	glGetProgramInfoLog(programHandle_, sizeof(str), NULL, str);
	addToLog(str);

	if (!result)
  		throw compile_error( "Shader link error", getLog(), programText );
  		
	CHECK_GL_DEBUG;
}

GLProgram::pGLShader
GLProgram::initShader( pGLShader shader, internal::Shader::ShaderType type,
	std::string source ) {
	char str[4096]; // For GL-state reporting
	if ( !source.empty() ) {
		shader.reset( new internal::GLShader( type ) );
		if ( shader->read( source ) ) {
			glAttachShader( programHandle_, shader->getShaderID() );
		} else {
      		glGetShaderInfoLog( shader->getShaderID(), sizeof(str), NULL, str);
      		addToLog(str);
      		throw compile_error( "Shader compile error", getLog(), source );
		}
	} 
	CHECK_GL_DEBUG;
	
	return shader;
}

void GLProgram::setInputTexture(const char *name,
				boost::shared_ptr<Texture> texture)
{
   const GLint location = validate_texture(name, texture);
   unsigned int i;
   for(i=0; i<textureInfo_.size(); i++)
   {
      if(textureInfo_[i].name==name)
      {
	 textureInfo_[i].texture=texture;
	 break;
      }
   }

   if(i==textureInfo_.size())
   {
    TextureInfo info;
    info.texture=texture;
    info.name=name;
    textureInfo_.push_back(info);
  }
  glUseProgram(programHandle_);
  glUniform1i(location, i);
  CHECK_GL_DEBUG;
}

void GLProgram::setInputTexture(const char *name,
			   boost::shared_ptr<Texture> texture,
			   unsigned int texUnit)
{
   const GLint location = validate_texture(name, texture);

   unsigned int i;
   int first_empty=-1;
   for(i=0; i<textureInfo_.size(); i++)
   {
      if(textureInfo_[i].name==name)
      {
	 if (i==texUnit)
	 {
	    textureInfo_[i].texture=texture;
	    break;
	 } else
	 {
	    first_empty=i;
	    textureInfo_[i].texture.reset();
	    textureInfo_[i].name="";
	 }
      }
      if(first_empty<0 && textureInfo_[i].name=="" )
      {
	 first_empty=i;
      }
   }

   if(i==textureInfo_.size())
   {
      if (texUnit<textureInfo_.size() && textureInfo_[texUnit].name != "")
      {
	 // The texture unit is already in use, move the texture
	 if (first_empty<0)
	 {
	    textureInfo_.push_back(textureInfo_[texUnit]);
	 } else
	 {
	    textureInfo_[first_empty]=textureInfo_[texUnit];
	 }
	 textureInfo_[texUnit].texture=texture;
	 textureInfo_[texUnit].name=name;
      } else // The texture unit is not in use
      {
	 TextureInfo new_info;
	 new_info.texture=texture;
	 new_info.name=name;
      
	 if (textureInfo_.size()<texUnit+1)
	 {
	    textureInfo_.resize(texUnit+1);
	 }
	 textureInfo_[texUnit]=new_info;
      }
  }
  glUseProgram(programHandle_);
  glUniform1i(location, i);
  CHECK_GL_DEBUG;
}

void 
GLProgram::activate()
{
  CHECK_GL_DEBUG;
   activateRenderTargets();
  CHECK_GL_DEBUG;
  for (int i=0; i<(int)(textureInfo_.size()); i++)
  {
    glActiveTextureARB((GLenum) (GL_TEXTURE0_ARB + i));
    textureInfo_[i].texture->bind();
  }
  glActiveTextureARB((GLenum) (GL_TEXTURE0_ARB));
  CHECK_GL_DEBUG;
  frameBuffer_->validateCompleteness();
  CHECK_GL_DEBUG;
  glUseProgram(programHandle_);
}

void 
GLProgram::run()
{
   activate();
   const OffScreenBuffer *off_buf = 
      dynamic_cast<OffScreenBuffer *>(frameBuffer_.get());
   if (off_buf != 0)
   {
      int width, height;
      glPushAttrib(GL_VIEWPORT_BIT);
      off_buf->getSize(width, height);
      glViewport(0, 0, width, height);
   }      
   CHECK_GL_DEBUG;
  
   renderQuad();
   CHECK_GL_DEBUG;
   if (off_buf != 0)
   {
      glPopAttrib();
   }      
   CHECK_GL_DEBUG;
}



void GLProgram::setParam1f(const char *name, const float param) 
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
	glUniform1f(location, param);
	CHECK_GL_DEBUG;
}

void GLProgram::setParam1i(const char *name, const int param)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
	glUniform1i(location, param);
	CHECK_GL_DEBUG;
}

void GLProgram::setParam1fv(const char *name, const int count, const float *params)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform1fv(location, count, params);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam1iv(const char *name, const int count, const int *params)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform1iv(location, count, params);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam2f(const char *name, const float param0, const float param1)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform2f(location, param0, param1);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam2i(const char *name, const int param0, const int param1)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform2i(location, param0, param1);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam2fv(const char *name, const int count, const float *params)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform2fv(location, count, params);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam2iv(const char *name, const int count, const int *params)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform2iv(location, count, params);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam3f(const char *name, const float param0, const float param1, const float param2)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform3f(location, param0, param1, param2);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam3i(const char *name, const int param0, const int param1, const int param2)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform3i(location, param0, param1, param2);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam3fv(const char *name, const int count, const float *params)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
	glUniform3fv(location, count, params);
	CHECK_GL_DEBUG;
}

void GLProgram::setParam3iv(const char *name, const int count, const int *params)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
	glUniform3iv(location, count, params);
	CHECK_GL_DEBUG;
}

void GLProgram::setParam4f(const char *name, const float param0, const float param1, const float param2, const float param3)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform4f(location, param0, param1, param2, param3);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam4i(const char *name, const int param0, const int param1, const int param2, const int param3)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
  	glUniform4i(location, param0, param1, param2, param3);
  	CHECK_GL_DEBUG;
}

void GLProgram::setParam4fv(const char *name, const int count, const float *params)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
	glUniform4fv(location, count, params);
	CHECK_GL_DEBUG;
}

void GLProgram::setParam4iv(const char *name, const int count, const int *params)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
	glUniform4iv(location, count, params);
	CHECK_GL_DEBUG;
}

void GLProgram::setParamMatrix4fv(const char *name, const float *params)
{
	int location = validate_uniform( name );
	glUseProgram(programHandle_);
	glUniformMatrix4fv(location, 1, false, params);
	CHECK_GL_DEBUG;
}

void GLProgram::setAttrib1f(const char *name, const float attrib)
{
  int location = validate_attrib( name );
  glVertexAttrib1f(location, attrib);
  CHECK_GL_DEBUG;
}

void GLProgram::setAttrib2f(const char *name, const float attrib0, const float attrib1)
{
  int location = validate_attrib( name );
  glVertexAttrib2f(location, attrib0, attrib1);
  CHECK_GL_DEBUG;
}

void GLProgram::setAttrib2fv(const char *name, const float *attribs)
{
  int location = validate_attrib( name );
  glVertexAttrib2fv(location, attribs);
  CHECK_GL_DEBUG;
}

void GLProgram::setAttrib3f(const char *name, const float attrib0, const float attrib1, const float attrib2)
{
  int location = validate_attrib( name );
  glVertexAttrib3f(location, attrib0, attrib1, attrib2);
  CHECK_GL_DEBUG;
}

void GLProgram::setAttrib3fv(const char *name, const float *attribs)
{
  int location = validate_attrib( name );
  glVertexAttrib3fv(location, attribs);
  CHECK_GL_DEBUG;
}
void GLProgram::setAttrib4f(const char *name, const float attrib0, const float attrib1, const float attrib2, const float attrib3)
{
  int location = validate_attrib( name );
  glVertexAttrib4f(location, attrib0, attrib1, attrib2, attrib3);
  CHECK_GL_DEBUG;
}

void GLProgram::setAttrib4fv(const char *name, const float *attribs)
{
  int location = validate_attrib( name );
  glVertexAttrib4fv(location, attribs);
  CHECK_GL_DEBUG;
}

void GLProgram::setAttribPointer(const char *name, int size, int type, unsigned int stride, 
                                 const void *pointer, bool normalized)
{
  int location = validate_attrib( name );
  glBindAttribLocation(programHandle_, location, name);
  glVertexAttribPointer(location, size, type, normalized, stride, pointer);
  CHECK_GL_DEBUG;
}

void GLProgram::enableClientState(const char *name)
{
  int location = validate_attrib( name );
  glEnableVertexAttribArray(location);
  CHECK_GL_DEBUG;
}

void GLProgram::disableClientState(const char *name)
{
  int location = validate_attrib( name );
  glEnableVertexAttribArray(location);
  CHECK_GL_DEBUG;
}

int 
GLProgram::validate_texture(const char* name, 
		 boost::shared_ptr<Texture> texture ) 
   const throw (std::invalid_argument )
{
   const GLint location = glGetUniformLocation(programHandle_, name);
   if ( location == -1 ) {
      std::string s = "Uniform variable ";
      s += name;
      s += " was not found or illegal.\n(Hint: starting with gl_ is not allowed.)";
      throw std::invalid_argument( s );
   }
	
   NameTypeMap::const_iterator it = uniformsNameToType_.find( name );
   assert( it != uniformsNameToType_.end() && 
	   "Internal Shallows error. The uniform map does not correspond with OpenGL." );
	
   try {
      texture->validateSamplerType(it->second);
   } catch (std::invalid_argument &e)
   {
      std::string s = "Varying variable ";
      s += name;
      s += " ";
      s += e.what();
      throw std::invalid_argument( s );
   }
   return location;
}

GLint
GLProgram::validate_uniform( const char* name ) const throw (std::invalid_argument, shallows_runtime_error ) {
	const GLint location = glGetUniformLocation(programHandle_, name);
	if ( location == -1 ) {
		std::string s = "Uniform variable ";
		s += name;
		s += " was not found or illegal.\n(Hint: The compiler might have optimized it away.)\n";
		throw std::invalid_argument( s );
	}
	
	// Name could index into an array. We remove everything after
	// the first trailing '['
	string nm = name;
	string::size_type i = nm.find_first_of( '[' );
	nm = nm.substr( 0, i );
	
	NameTypeMap::const_iterator it = uniformsNameToType_.find( nm );
	if ( it == uniformsNameToType_.end() ) {
	    std::ostringstream out;
	    out << "Internal Shallows error. The uniform map does not correspond with OpenGL.\n";
	    //out << printUniformsMap();
	    out << std::endl;
	    throw shallows_runtime_error( out.str() );
	}

	std::string type = it->second;
	if ( (type.find( "sampler" ) != string::npos) ) {
		std::string s = "Uniform variable ";
		s += name;
		s += " is  a sampler (texture). It can not be set this way. ";
		throw std::invalid_argument( s );
	}
	
	return location;
}

GLint
GLProgram::validate_attrib( const char* name ) const throw (std::invalid_argument) {
	const GLint location = glGetAttribLocation( programHandle_, name );
	if ( location == -1 ) {
		std::string s = "Varying variable ";
		s += name;
		s += " was not found or illegal.\n(Hint starting with gl_ is not allowed.)";
		throw std::invalid_argument( s );
	}
	
	return location;
}
		
void
GLProgram::buildUniformsMap( const std::string shaderSource ) {
    uniformsNameToType_.clear();
    
    using std::string;
    
    // First remove all comments
    string src = boost::regex_replace( shaderSource, boost::regex( "(//[^\\n]*|/\\*.*?\\*/)" ), 
        "" );
    
    string::const_iterator start, end;
    start = src.begin();
    end = src.end();
    boost::match_results<std::string::const_iterator> what; 
    boost::match_flag_type flags = boost::match_default; 
    
  while(boost::regex_search( start, end, what, boost::regex( "uniform[[:space:]]([^;]*)[[:space:]]([^;]*)[;]" ), flags)) { 
      /*
       * what[0] is the full match
       * what[1] is the type
       * what[2] is the name
       */
      
      // Move the start of the shader forward.
      start = what[0].second; 
      
      // The type could be an array (of arrays of arrays...), so
      // we strip everything from the first `[` and backwards.
      string name = what[2];
      name = name.substr( 0, name.find_first_of( '[' ) );
      //std::cout << what[2] << " ,type: " << type << std::endl;
      // Add the uniform to the map
      uniformsNameToType_[name] = what[1];
      // update flags: 
      flags |= boost::match_prev_avail; 
      flags |= boost::match_not_bob; 
    }
}

unsigned int 
GLProgram::getProgramHandle() {
	return programHandle_;
}

} // namespace shallows
