 /**********************************************************************
 * 
 *  File: OffScreenBuffer.cpp
 * 
 *  Created:
 *
 *  Version: $Id: OffScreenBuffer.cpp,v 1.23 2005/12/14 17:23:09 trondrh Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/
#include "shallows/OffScreenBuffer.hpp"

#include "shallows/internal/GLIncludes.hpp"
#include "shallows/shallows.hpp"
#include "shallows/shallowsexcept.hpp"
#include "shallows/Texture2D.hpp"
#include "shallows/RenderTarget.hpp"
#include "shallows/RenderTexture2D.hpp"
#include "shallows/OffScreenRenderTarget.hpp"
#include <boost/weak_ptr.hpp>

using namespace shallows::internal;
using boost::shared_ptr;
using boost::weak_ptr;

namespace shallows {

OffScreenBuffer::OffScreenBuffer(unsigned int width, unsigned int height)  :
   height_(0), width_(0), internalFormat_(0), id_(0), depthFormat_(0)
{ 
   init(GL_RGBA_FLOAT32_ATI, width, height);
 }

OffScreenBuffer::OffScreenBuffer(int internalFormat, unsigned int width, 
				 unsigned int height)  :
   height_(0), width_(0), internalFormat_(internalFormat), id_(0), depthFormat_(0)
{
   init(internalFormat_, width, height);
}

OffScreenBuffer::OffScreenBuffer(Texture2D &t) :
   height_(0), width_(0), internalFormat_(0), id_(0)
{
   try {
      init(t.getInternalFormat(), t.getWidth(), t.getHeight());
   } catch ( ... )
   {
      height_=0;
      width_=0;
      internalFormat_=0;
      throw ;
   }
}

   
OffScreenBuffer::~OffScreenBuffer()
{
   if (id_>0)
      glDeleteFramebuffersEXT(1, &id_);
}

boost::shared_ptr<RenderBuffer>  
OffScreenBuffer::createRenderBuffer()
{
   boost::shared_ptr<RenderBuffer> rt;
   rt.reset(new RenderBuffer(internalFormat_, width_, height_));
   CHECK_GL_DEBUG;
   return rt;
}

boost::shared_ptr<RenderTexture2D> 
  OffScreenBuffer::createRenderTexture2D()
{
   boost::shared_ptr<Texture2D> tex;
   tex.reset(new Texture2D(internalFormat_,
			   width_,
			   height_,
			   GL_RGBA,
			   GL_FLOAT,
			   0));
   tex->bind();
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP); 
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

   boost::shared_ptr<RenderTexture2D> rt(new RenderTexture2D(tex));
   CHECK_GL_DEBUG;
   return rt;
}

boost::shared_ptr<RenderTexture2D> 
  OffScreenBuffer::createRenderTextureRectangle()
{
   boost::shared_ptr<Texture2D> tex;
   tex.reset(new Texture2D(internalFormat_,
			   width_,
			   height_,
			   GL_RGBA,
			   GL_FLOAT,
			   0,
			   true));
   tex->bind();
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP); 
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, 
		    GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, 
		    GL_TEXTURE_MAG_FILTER, GL_NEAREST);

   boost::shared_ptr<RenderTexture2D> rt(new RenderTexture2D(tex));
   CHECK_GL_DEBUG;
   return rt;
}

boost::shared_ptr<RenderBuffer> 
OffScreenBuffer::createDepthBuffer()
{
   if (depthFormat_==0)
     depthFormat_=GL_DEPTH_COMPONENT24;
   
   boost::shared_ptr<RenderBuffer> rt;
   rt.reset(new RenderBuffer(depthFormat_, width_, height_));
   CHECK_GL_DEBUG;
   return rt;
}

boost::shared_ptr<RenderTexture2D>  
OffScreenBuffer::createDepthTexture(bool rectangular)
{
   if (depthFormat_==0)
     depthFormat_=GL_DEPTH_COMPONENT24;
   boost::shared_ptr<Texture2D> tex;
   tex.reset(new Texture2D(depthFormat_,
			   width_,
			   height_,
         GL_DEPTH_COMPONENT,
			   GL_FLOAT,
			   0,
			   rectangular));
   tex->bind();
   GLenum target = rectangular?GL_TEXTURE_RECTANGLE_ARB:GL_TEXTURE_2D;
   glTexParameteri( target, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri( target, GL_TEXTURE_WRAP_T, GL_CLAMP); 
   glTexParameteri( target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

   boost::shared_ptr<RenderTexture2D> rt(new RenderTexture2D(tex));
   CHECK_GL_DEBUG;
   return rt;
}

std::vector<boost::weak_ptr<OffScreenRenderTarget> > &
OffScreenBuffer::getTargets()
{
   return targets_;
}

unsigned int OffScreenBuffer::attachRenderTarget(
   shared_ptr<shallows::OffScreenRenderTarget> rt, int colAttachable)
{

   boost::shared_ptr<OffScreenRenderTarget> off_rt;
   try {
      off_rt= boost::static_pointer_cast<OffScreenRenderTarget>(rt);
   } catch ( ... ) 
   { 
      throw 
	 shallows_invalid_argument("Attempt to attach OnScreenRenderTarget to OffScreenBuffer");
   }

   unsigned int drawEnum=attach(this, off_rt, colAttachable);
   glDrawBuffer(drawEnum);
   glReadBuffer(drawEnum);
   return drawEnum;
   CHECK_GL_DEBUG;
}

void OffScreenBuffer::attachRenderTarget(
   boost::shared_ptr<shallows::OffScreenRenderTarget> *rt, 
   unsigned int numBuffs)
{
   unsigned int enums[16];
   if (numBuffs>16)
   {
     throw 
       shallows_invalid_argument("attachRenderTarget called with more than 16 buffers");
   }
   
   attach(this, rt, enums, numBuffs);
   glDrawBuffers(numBuffs, enums);
   glReadBuffer(GL_NONE);
} 

void OffScreenBuffer::detachRenderTarget(int att_point)
{
   glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
				GL_COLOR_ATTACHMENT0_EXT+att_point,
				GL_RENDERBUFFER_EXT, 0);
   targets_[att_point].reset();
}
/*
void OffScreenBuffer::attachRenderTarget(
RenderTarget &r, int colBufId)
{
   OffScreenRenderTarget &of_rt = dynamic_cast<OffScreenRenderTarget &> (r);
   if (&of_rt == 0)
   {
      throw 
	 shallows_invalid_argument("Attempt to attach OnScreenRenderTarget to OffScreenBuffer");
   } 
      
   validateRenderTarget(of_rt);
   activate();
   of_rt.bindToCurrentFrameBuffer(colBufId);
   CHECK_GL_DEBUG;
}
*/

void 
OffScreenBuffer::attachDepthBuffer(RenderTarget &r)
{
   OffScreenRenderTarget &of_rt = dynamic_cast<OffScreenRenderTarget &> (r);
   if (&of_rt == 0)
   {
      throw 
	 shallows_invalid_argument("Attempt to attach OnScreenRenderTarget to OffScreenBuffer");
   } 
      
   activate();
   of_rt.bindAsDepthToCurrentFrameBuffer();
   CHECK_GL_DEBUG;
}


void OffScreenBuffer::validateRenderTarget(const RenderTarget &r) const
{
   const OffScreenRenderTarget &of_rt = 
      dynamic_cast<const OffScreenRenderTarget &> (r);
   if (&of_rt == 0)
   {
      throw 
	 shallows_invalid_argument("OnScreenRenderTarget does not match OffScreenBuffer");
   } 
      
   validateRenderTarget(of_rt);
}

void OffScreenBuffer::validateRenderTarget(const OffScreenRenderTarget &r) const
{
   if (r.getHeight() != height_)
   {
      throw 
	 shallows_invalid_argument("RenderTarget height does not match frame buffer");
   } 
   if (r.getWidth() != width_) 
   {
      throw 
	 shallows_invalid_argument("RenderTarget width does not match frame buffer");
   } 
   if (r.getInternalFormat() != internalFormat_)
   {
      throw 
	 shallows_invalid_argument("Texture type does not match frame buffer");
   } 
}

void OffScreenBuffer::validateCompleteness() const
{
   CHECK_GL_DEBUG;
   glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, id_);
   GLenum status;
   status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
   if (status==GL_FRAMEBUFFER_COMPLETE_EXT)
      return;
   std::string s;
   s = "FrameBuffer incomplete, ";
   switch(status) {
   case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
      s+= "GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT";
      break;
   case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
      s+= "GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT";
      break;
   case GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT:
      s+= "GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT";
      break;
   case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
      s+= "GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT";
      break;
   case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
      s+= "GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT";
      break;
   case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
      s+= "GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT";
      break;
   case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
      s+= "GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT";
      break;
   case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
      s+= "GL_FRAMEBUFFER_UNSUPPORTED_EXT";
      break;
   }
   throw frame_buffer_incomplete(s, status);
}

void OffScreenBuffer::getSize(int &width, int &height) const
{
   width=width_;
   height=height_;
}

void OffScreenBuffer::init(int internalFormat, unsigned int width, 
			   unsigned int height)
{
   if (height<=0)
   {
      throw 
	 shallows_invalid_argument("OffScreenBuffer::OffScreenBuffer called with invalid height");
      
   }
   if (width<=0)
   {
      throw 
	 shallows_invalid_argument("OffScreenBuffer::OffScreenBuffer called with invalid width");
      
   }
   height_=height;
   width_=width;
   internalFormat_=internalFormat;

   glGenFramebuffersEXT(1, &id_);
   CHECK_GL_DEBUG;
   int max_draw_buffers;
   glGetIntegerv(GL_MAX_DRAW_BUFFERS, &max_draw_buffers);
   int max_color_attachments;
   glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT, &max_color_attachments);
   if (max_draw_buffers>max_color_attachments)
      max_draw_buffers=max_color_attachments;
   drawBuffers_.resize(max_draw_buffers);
   targets_.resize(max_draw_buffers);
   for (unsigned int i=0; i<(unsigned int)(max_draw_buffers); i++)
   {
      drawBuffers_[i]=static_cast<unsigned int> (GL_COLOR_ATTACHMENT0_EXT)+i;
   }
}

void OffScreenBuffer::setInternalFormat(Texture2D &t)
{
   try {
   height_=t.getHeight();
   width_=t.getWidth();
   internalFormat_=t.getInternalFormat();
   } catch ( ... )
   {
      height_=0;
      width_=0;
      internalFormat_=0;
      throw ;
   }
   CHECK_GL_DEBUG;
}


void OffScreenBuffer::operator=(const OffScreenBuffer &b)
{ ; }

void OffScreenBuffer::activate()
{
   if (id_==0)
   {
      throw
	 std::logic_error("Frame buffer failed to initialize");
   }

   CHECK_GL_DEBUG;
   glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, id_);
   CHECK_GL_DEBUG;
}

void  OffScreenBuffer::setRenderTargetAttached(
   boost::shared_ptr<shallows::OffScreenRenderTarget> &rt, 
   int attachment_point)
{
   targets_[attachment_point]=rt;

}

unsigned int attach(OffScreenBuffer *off_buff, 
	    boost::shared_ptr<shallows::OffScreenRenderTarget> &rt, 
	    int colAttachable)
{
   std::vector<boost::weak_ptr<OffScreenRenderTarget> > &targets=
      off_buff->getTargets();
   boost::shared_ptr<OffScreenRenderTarget> off_rt;
   try {
      off_rt= boost::static_pointer_cast<OffScreenRenderTarget>(rt);
   } catch ( ... ) 
   { 
      throw 
	 shallows_invalid_argument("Attempt to attach OnScreenRenderTarget to OffScreenBuffer");
   }

   off_buff->validateRenderTarget(*off_rt);
   off_buff->activate();
   
   int att_point=0;
   if (colAttachable<0 || targets[colAttachable].lock()!=off_rt)
   {
      for (int i=targets.size()-1; i>=0; i--)
      {
	 if (targets[i].expired())
	 {
	    att_point=i;
	    continue;
	 }
	 if (targets[i].lock()==off_rt)
	 {
	    att_point=i;
	    if (colAttachable>=0) // deactivate the previous point
	    {
	       off_buff->detachRenderTarget(att_point);
	       att_point=colAttachable;
	    } 
	    break;
	 }
      }
   }

   if (colAttachable>=0)
      att_point=colAttachable;
   boost::shared_ptr<OffScreenRenderTarget> previous=targets[att_point].lock();
   if (previous!=off_rt)
   {
      off_buff->setRenderTargetAttached(rt, att_point);
      rt->attachToCurrentFrameBuffer(att_point);
   }
   CHECK_GL_DEBUG;
   return GL_COLOR_ATTACHMENT0_EXT+att_point;
}

void attach(OffScreenBuffer *off_buff, 
	    boost::shared_ptr<shallows::OffScreenRenderTarget> *rt, 
	    unsigned int *enums,
	    unsigned int numBuffs)
{
   off_buff->activate();
   const int num_att_points=off_buff->targets_.size();
   int availible[16];
   int used[16];
   int decided[16];

   if (num_att_points>16) 
   { 
     throw 
       shallows_invalid_argument("Attach called with more than 16 buffers");
   }

   unsigned int n_availible=0;
   unsigned int n_used=0;

   std::vector<boost::weak_ptr<OffScreenRenderTarget> > &targets=
      off_buff->getTargets();

   for (int i=0; i<numBuffs; i++)
   {
      decided[i]=-1;
   }
   for (int i=num_att_points-1; i>=0; i--)
   {
      if (targets[i].expired())
	 availible[n_availible++]=i;
      else 
	 used[n_used++]=i;
   }
   for (int i=0; i<numBuffs; i++) // Check if already bound
   {
      boost::shared_ptr<OffScreenRenderTarget> off_rt;
      try {
	 off_rt= boost::static_pointer_cast<OffScreenRenderTarget>(rt[i]);
      } catch ( ... ) 
      { 
	 throw 
	    shallows_invalid_argument("Attempt to attach OnScreenRenderTarget to OffScreenBuffer");
      }
      for (int j=0; j<n_used; j++)
      {
	 if (targets[used[j]].lock()==off_rt)
	 {
	    decided[i]=used[j];
	    used[j]=used[--n_used];
	    break;
	 }
      }
   }   
   for (int i=0; i<numBuffs; i++) // pick unused if possible
   {
      if (decided[i]>=0)
	 continue;
      if (n_availible>0)
	 decided[i]=availible[--n_availible];
      else
	 decided[i]=used[--n_used];
      
      off_buff->setRenderTargetAttached(rt[i], decided[i]);
      rt[i]->attachToCurrentFrameBuffer(decided[i]);
   }

   for (int i=0; i<numBuffs; i++) // pick unused if possible
   {
      enums[i]=GL_COLOR_ATTACHMENT0_EXT+decided[i]; 
   }
}

} // namespace shallows
