/**********************************************************************
 * 
 *	File: Program.cpp
 * 
 * 	Created: 20050621
 *
 *  Version: $Id: Program.cpp,v 1.27 2006/02/03 09:15:07 josse77 Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *			 Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *			 Johan Simon Seland <johans@ifi.uio.no>
 *
 * 	This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 * 	This library is free software; you can redistribute it and/or
 * 	modify it under the terms of the GNU General Public License
 * 	("GPL") version 2 as published by the Free Software Foundation.
 * 	See the file LICENSE.GPL at the root directory of this source
 * 	distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#include "shallows/Program.hpp"
#include "shallows/FrameBuffer.hpp"
#include "shallows/shallowsexcept.hpp"
#include "boost/regex.hpp"
#include "shallows/RenderTarget.hpp"
#include "shallows/internal/GLIncludes.hpp"
#include "shallows/shallows.hpp"
#include <fstream>
#include <string>
#include <iostream>
#include <sstream>

using namespace shallows::internal;

namespace shallows {
	
Program::Program() :
   uniforms( uniformsNameToType_ ), useViewportTexCoords_(true)
{
   int max_draw_buffers, max_color_attachments;
   glGetIntegerv(GL_MAX_DRAW_BUFFERS, &max_draw_buffers);
   glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT, &max_color_attachments);
   //   renderTargets_.resize(max_draw_buffers);
   renderTarget_.resize(max_draw_buffers);
}

Program::~Program() {
}

void
Program::readFile( const char *filename, const char *defines ) {
	std::ifstream file( filename );
	
	if ( !file ) {
		std::string s = filename;
		s += " was not found.";
		throw file_not_found( s );
	}
	 
	std::istreambuf_iterator<char> begin(file), end;
    std::string src(begin, end);	
    
    std::string str_def;
    if (defines)
       str_def=std::string(defines);
    read(src, str_def);
}

void
Program::read( std::string programText ) {
	std::string s;
  	read(programText, s);
}

std::string
Program::getLog() const
{
  return log_;
}

void 
Program::clearLog()
{
  log_.clear();
}

	
std::string
Program::getVertexShaderSource( const std::string programText, const std::string defines ) {
	boost::match_results<std::string::const_iterator> result; 
	/* In a programText we allow both the [Vertex shader] and the [Fragment shader]
	 * to come first.
	 */
	if ( boost::regex_search( programText, result, boost::regex( "\\[Vertex shader\\](.*)\\[Fragment shader\\]" ) ) )
    return defines + static_cast<std::string>(result[1]);
	else if ( boost::regex_search( programText, result,  boost::regex( "\\[Vertex shader\\](.*)" ) ) )
    return defines + static_cast<std::string>(result[1]);
	else
		return defines;
}

std::string
Program::getFragmentShaderSource( const std::string programText, const std::string defines ) {
	boost::match_results<std::string::const_iterator> result; 
	/* In a programText we allow both the [Vertex shader] and the [Fragment shader]
	 * to come first.
	 */
	if ( boost::regex_search( programText, result, boost::regex( "\\[Fragment shader\\](.*)\\[Vertex shader\\]" ) ) )
		return defines + static_cast<std::string>(result[1]);
	else if ( boost::regex_search( programText, result,  boost::regex( "\\[Fragment shader\\](.*)" ) ) )
		return defines + static_cast<std::string>(result[1]);
	else
		return defines;
}

void 
Program::addToLog(const std::string &s)
{
  log_.append(s);
}

void 
Program::setFrameBuffer(boost::shared_ptr<FrameBuffer> fb)
{
   frameBuffer_=fb;
}

void
Program::setOutputTarget( unsigned int colorNum, boost::shared_ptr<RenderTarget> target )
{
  CHECK_GL_DEBUG;
   if (colorNum >= renderTarget_.size() || colorNum<0)
   {
      std::string s = "Invalid colorNum used in Program::setOutputTarget ";
      s+= colorNum;
      s+= " not in range [0..";
      s+= (char)(renderTarget_.size())-1;
      s+= "]";
      throw std::invalid_argument( s );
   }
  CHECK_GL_DEBUG;
   if(frameBuffer_.get() == 0)
   { 
      throw std::invalid_argument( "cannot attach rendertarget before framebuffer");
   }
  CHECK_GL_DEBUG;
   frameBuffer_->validateRenderTarget(*target.get());
   renderTarget_[colorNum]=target;
   CHECK_GL_DEBUG;
}


void 
Program::activateRenderTargets()
{
   //We allow the user to use a program without specifying frameBuffer
   if (frameBuffer_.get() == 0)
      return; 
   int numRenderTargets=0;
   frameBuffer_->activate();
   OffScreenBuffer *off_buf = 
      dynamic_cast<OffScreenBuffer *>(frameBuffer_.get());

   for (int i=0; i<(int)(renderTarget_.size()); i++)
   {
      if (renderTarget_[i].get()!=0)
      {
	 if (numRenderTargets!= i)
	 {
	    throw std::runtime_error("Missing rendertargets");
	 }
	 numRenderTargets++;
      }
   } 
      
   if (off_buf!=0)
   {
     boost::shared_ptr<shallows::OffScreenRenderTarget>  rt[16];
//     unsigned int enums[16];
      for (int i=0; i<numRenderTargets; i++)
	 rt[i]=boost::dynamic_pointer_cast<OffScreenRenderTarget>
	    (renderTarget_[i]);

      off_buf->attachRenderTarget(rt, numRenderTargets);
//      attach(off_buf, rt, enums, numRenderTargets);
//      glDrawBuffers(numRenderTargets, enums);
   }
   const OnScreenBuffer *on_buf = 
      dynamic_cast<OnScreenBuffer *>(frameBuffer_.get());
   if (on_buf  != 0)
   {
      if (numRenderTargets > 1)
      {
	 throw not_implemented("MRT not implemented for on screen buffers");
      }
      const OnScreenRenderTarget *on_rt = 
	 dynamic_cast<OnScreenRenderTarget *>
	 (renderTarget_[0].get());
      glDrawBuffer(on_rt->getLastBufferEnum());
   }
}





void 
Program::renderQuad()
{
   if (useViewportTexCoords_)
   {
      int viewport[4];
      glGetIntegerv(GL_VIEWPORT, viewport);
      min_u_=0.0f;
      min_v_=0.0f;
      max_u_=(float)(viewport[2]);
      max_v_=(float)(viewport[3]);
  }
   
   glBegin(GL_QUADS);
   glTexCoord2f(min_u_, min_v_);
   glVertex3f(-1.0f, -1.0f, 0.5f);

   glTexCoord2f(max_u_, min_v_);
   glVertex3f(1.0f, -1.0f, 0.5f);

   glTexCoord2f(max_u_, max_v_);
   glVertex3f(1.0f, 1.0f, 0.5f);

   glTexCoord2f(min_u_, max_v_);
   glVertex3f(-1.0f, 1.0f, 0.5f);
   glEnd();
}     

void 
Program::getTexCoords(float &min_u, float &min_v, 
		      float &max_u, float &max_v) const
{
   min_u=min_u_;
   min_v=min_v_;
   max_u=max_u_;
   max_v=max_v_;
}

void 
Program::setTexCoords(float min_u, float min_v, 
		      float max_u, float max_v)
{
   min_u_=min_u;
   min_v_=min_v;
   max_u_=max_u;
   max_v_=max_v;
   useViewportTexCoords_=false;
}

void 
Program::useViewportTexCoords()
{
   const OffScreenBuffer *off_buf = 
      dynamic_cast<OffScreenBuffer *>(frameBuffer_.get());
   int viewport[4];
   if (off_buf != 0)
   {
      viewport[0]=0;
      viewport[1]=0;
      off_buf->getSize(viewport[2], viewport[3]);
   } else
   {
      glGetIntegerv(GL_VIEWPORT, viewport);
   }
   min_u_=(float)(viewport[0]);
   min_v_=(float)(viewport[1]);
   max_u_=(float)(viewport[2]);
   max_v_=(float)(viewport[3]);
   useViewportTexCoords_=true;
}

void 
Program::useNormalizedTexCoords()
{
   min_u_=0.0f;
   min_v_=0.0f;
   max_u_=1.0f;
   max_v_=1.0f;
   useViewportTexCoords_=false;
}

std::string
Program::printUniformsMap() const
{
    std::ostringstream out;

    NameTypeMap::const_iterator i = uniformsNameToType_.begin();
    while ( i != uniformsNameToType_.end() ) {
	out << i->first << " --> " << i->second << std::endl;
	i++;
    }

    return out.str();
    
}
} // namespace shallows
