/**********************************************************************
 * 
 *  File: textureCreation.cpp
 * 
 *  Created: 11. July 2005
 *
 *  Version: $Id: textureCreation.cpp,v 1.12 2005/10/26 12:35:30 dyken Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	         Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	         Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#include "shallows/utils/textureCreation.hpp"
#include <boost/shared_ptr.hpp>
#include "shallows/shallows.hpp"
#include "shallows/Texture2D.hpp"
#include "shallows/Texture1D.hpp"
#include "shallows/internal/GLIncludes.hpp"
#include "shallows/shallowsexcept.hpp"

using namespace shallows;
using boost::shared_ptr;

namespace shallows {
namespace utils {
boost::shared_ptr<Texture2D> createFloatTexture2D(unsigned int width, 
						  unsigned int height,
						  unsigned int components,
						  const float *pixels)
{
   GLenum internalformat;
   GLenum format; 
   if (components == 1)
   {
      internalformat = GL_FLOAT_R32_NV;
      format = GL_RED;
      //      internalformat=GL_ALPHA32F_ARB;
      //format = GL_ALPHA;
   }
   else if (components == 2)
   {
      internalformat = GL_LUMINANCE_ALPHA32F_ARB;
      format = GL_LUMINANCE_ALPHA;
   }
   else if (components == 3)
   {
     internalformat=GL_RGB32F_ARB;
     format=GL_RGB;
   }
   else if (components == 4)
   {
     internalformat=GL_RGBA32F_ARB;
     format=GL_RGBA;
   }
   else
     throw std::invalid_argument
       ( "readFloatBuffer called with invalid number of components" );

   boost::shared_ptr<Texture2D> tex;
   tex.reset(new shallows::Texture2D (internalformat,
				      width, height, 
				      format,
				      GL_FLOAT, 
				      pixels)); 
   tex->bind();
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP); 
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   
   return tex;
}

boost::shared_ptr<Texture1D> createFloatTexture1D(unsigned int width, 
						  unsigned int components,
						  const float *pixels)
{
   GLenum internalformat;
   GLenum format;
   if (components == 1)
   {
      internalformat = GL_FLOAT_R32_NV;
      format = GL_RED;
      //      internalformat=GL_ALPHA32F_ARB;
      //format = GL_ALPHA;
   }
   else if (components == 2)
   {
      internalformat = GL_LUMINANCE_ALPHA32F_ARB;
      format = GL_LUMINANCE_ALPHA;
   }
   else if (components == 3)
   {
     internalformat=GL_RGB32F_ARB;
     format=GL_RGB;
   }
   else if (components == 4)
   {
     internalformat=GL_RGBA32F_ARB;
     format=GL_RGBA;
   }
   else
     throw std::invalid_argument
       ( "readFloatBuffer called with invalid number of components" );

   boost::shared_ptr<Texture1D> tex;
   tex.reset(new shallows::Texture1D (internalformat,
				      width, 
				      format,
				      GL_FLOAT, 
				      pixels)); 
   tex->bind();
   glTexParameteri( GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri( GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, GL_CLAMP); 
   glTexParameteri( GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   
   return tex;
}


shared_ptr<Texture2D> createFloatTextureRectangle(unsigned int width, 
							 unsigned int height, 
							 unsigned int components,
							 const float *pixels)
{
   GLenum internalformat;
   GLenum format;
   if (components == 1)
   {
      internalformat = GL_FLOAT_R32_NV;
      format = GL_RED;
      //     internalformat=GL_ALPHA32F_ARB;
      //     format = GL_ALPHA;
   }
   else if (components == 2)
   {
      internalformat = GL_FLOAT_RG32_NV;
      format = GL_RED;
      //      internalformat = GL_LUMINANCE_ALPHA32F_ARB;
      //format = GL_LUMINANCE_ALPHA;
   }
   else if (components == 3)
   {
     internalformat=GL_RGB32F_ARB;
     format=GL_RGB;
   }
   else if (components == 4)
   {
     internalformat=GL_RGBA32F_ARB;
     format=GL_RGBA;
   }
   else
     throw std::invalid_argument
       ( "readFloatBuffer called with invalid nunber of components" );

   boost::shared_ptr<Texture2D> tex;
   tex.reset(new shallows::Texture2D (internalformat,
				      width, height, 
				      format,
				      GL_FLOAT, 
				      pixels,
				      true)); 
   tex->bind();
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP); 
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   
   return tex;
}


shared_ptr<Texture2D> createUnsignedByteTextureRectangle
  (unsigned int width, 
   unsigned int height, 
   unsigned int components,
   const unsigned char *pixels)
{
   GLenum internalformat;
   GLenum format;
//     if (components == 1)
//     {
//       internalformat=GL_ALPHA8;
//       format = GL_ALPHA;
//     }
//     else if (components == 2)
//     {
//        internalformat = GL_LUMINANCE8_ALPHA8;
//        format = GL_LUMINANCE_ALPHA;
//     }
//   else
   if (components == 3)
   {
     internalformat=GL_RGB8;
     format=GL_RGB;
   }
   else if (components == 4)
   {
     internalformat=GL_RGBA8;
     format=GL_RGBA;
   }
   else
     throw std::invalid_argument
       ( "createUnsignedByteTextureRectangle called with invalid number of components" );

   boost::shared_ptr<Texture2D> tex;
   tex.reset(new shallows::Texture2D (internalformat,
				      width, height, 
				      format,
				      GL_UNSIGNED_BYTE, 
				      pixels,
				      true)); 
   tex->bind();
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP); 
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   
   return tex;
}

boost::shared_ptr<Texture2D> createUnsignedByteTexture2D(unsigned int width, 
						  unsigned int height,
						  unsigned int components,
						  const unsigned char *pixels)
{
   GLenum internalformat;
   GLenum format;
//     if (components == 1)
//     {
//       internalformat=GL_ALPHA8;
//       format = GL_ALPHA;
//     }
//     else if (components == 2)
//     {
//        internalformat = GL_LUMINANCE8_ALPHA8;
//        format = GL_LUMINANCE_ALPHA;
//     }
//     else 
   if (components == 3)
   {
     internalformat=GL_RGB8;
     format=GL_RGB;
   }
   else if (components == 4)
   {
     internalformat=GL_RGBA8;
     format=GL_RGBA;
   }
   else
     throw std::invalid_argument
       ( "createUnsignedByteTexture2D called with invalid number of components" );

   boost::shared_ptr<Texture2D> tex;
   tex.reset(new shallows::Texture2D (internalformat,
				      width, height, 
				      format,
				      GL_UNSIGNED_BYTE, 
				      pixels)); 
   tex->bind();
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP); 
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   
   return tex;
}

boost::shared_ptr<Texture3D> createFloatTexture3D(unsigned int width, 
												  unsigned int height,
												  unsigned int depth,
												  unsigned int components,
												  const float *pixels)
{
	if((width%2) != 0) throw std::invalid_argument
		("createFloatTexture3D called with width not multiple of two");

	if((height%2) != 0) throw std::invalid_argument
		("createFloatTexture3D called with height not multiple of two");

	if((depth%2) != 0) throw std::invalid_argument
		("createFloatTexture3D called with depth not multiple of two");

	GLenum internalformat;
	GLenum format;

	switch(components) {
	case 4:
		internalformat = GL_RGBA_FLOAT32_ATI; 
		format = GL_RGBA;
		break;
	default:
		throw std::invalid_argument
			("createFloatTexture3D called with invalid number of components" );		
	}
	
	boost::shared_ptr<Texture3D> tex;
	tex.reset(new shallows::Texture3D(internalformat,
									  width,
									  height,
									  depth,
									  format,
									  GL_FLOAT,
									  pixels));
	tex->bind();
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	return tex;
}


} // namespace shallows::utils
} //namespace utils 

