/**********************************************************************
 * 
 *  File: CgProgramTest.cpp
 * 
 *  Created: Jul 11, 2005
 *
 *  Version: $Id: CgProgramTest.cpp,v 1.7 2005/09/15 09:12:23 jonhjelm Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/
#if defined(HAVE_CG) && defined(HAVE_CGGL)

#include "CgProgramTest.hpp"
#include "shallows/CgProgram.hpp"
#include "shallows/Texture2D.hpp"
#include "shallows/shallowsexcept.hpp"
#include "shallows/utils/textureCreation.hpp"
#include "shallows/shallows.hpp"
#include "shallows/utils/readBuffer.hpp"
#include "shallowstesting.hpp"
#include <string>
#include <cassert> 
#include <iostream>
#include <stdexcept>

using namespace std;
using boost::shared_ptr;
namespace test
{

CPPUNIT_TEST_SUITE_REGISTRATION( CgProgramTest );

CgProgramTest::CgProgramTest()
{
}

CgProgramTest::~CgProgramTest()
{
}

void
CgProgramTest::setUp() {
	initGlut();
}

void
CgProgramTest::construction() {
	CPPUNIT_ASSERT_NO_THROW( shallows::CgProgram P );	
}

void
CgProgramTest::read() {
	// Check that we correctly read empty shader strings.
	string s = "[Vertex shader]\nvoid main() { }";
	string vs = "\nvoid main() { }";
	CPPUNIT_ASSERT( shallows::Program::getFragmentShaderSource( s ).empty() );
	string vss = shallows::Program::getVertexShaderSource( s );
	CPPUNIT_ASSERT_EQUAL( vs, vss );
	
	s = "[Fragment shader]\nvoid main() {}";
	string fs = "\nvoid main() {}";
	CPPUNIT_ASSERT( shallows::Program::getVertexShaderSource( s ).empty() );
	string fss = shallows::Program::getFragmentShaderSource( s );
	CPPUNIT_ASSERT_EQUAL( fss, fs );
    
    // Check that we throw if we do not find the shader source
    shallows::CgProgram P;
    CPPUNIT_ASSERT_THROW( P.readFile( "bogusshader" ), shallows::file_not_found );
}

void
CgProgramTest::setParam() {
	shallows::CgProgram P;
	P.readFile( "shaders/showTex.cgshader" );
    
    shallows::Program::NameTypeMap::const_iterator it = P.uniforms.begin();
    //    for ( ; it != P.uniforms.end(); ++it )
    //    cout << it->first << " -> " << it->second << endl;
        
	CPPUNIT_ASSERT_THROW( P.setParam1f( "foo", 1.0f ), std::invalid_argument );
	
    CPPUNIT_ASSERT_THROW( P.setParam1f( "texture", 1.0f ), std::invalid_argument );
	
	shallows::CgProgram Q;
	Q.readFile( "shaders/attribshader.cgshader" );
	CPPUNIT_ASSERT_THROW( Q.setParam1f( "texture", 1.0f ), std::invalid_argument );
	CPPUNIT_ASSERT_THROW( Q.setParam1f( "color", 1.0f ), std::invalid_argument );
	CPPUNIT_ASSERT_THROW( Q.setAttrib4f( "foo", 0.0f, 0.0f, 0.0f, 0.0f ), std::invalid_argument );
	CPPUNIT_ASSERT_NO_THROW( Q.setAttrib4f( "IN.vertexdata", 0.0f, 0.0f, 0.0f, 0.0f ) );
    
}

void 
CgProgramTest::onlyFragmentShader() {
   using shallows::utils::createFloatTexture2D;
   using shallows::utils::createFloatTextureRectangle;
   shallows::CgProgram P;
   try { P.readFile( "shaders/frag_only.cgshader" ); }
   catch (shallows::compile_error &e)
   {
      std::cerr << e.log();
      throw;
   }
    
   boost::shared_ptr<shallows::Texture2D> tex;
   tex=createFloatTexture2D(4, 4, 4,0);

   boost::shared_ptr<shallows::Texture2D> texRect;
   texRect=createFloatTextureRectangle(4, 4, 4,0);
	
   CPPUNIT_ASSERT_NO_THROW( P.setInputTexture("texture", tex));
   CPPUNIT_ASSERT_NO_THROW( P.setInputTexture("textureRect", texRect));
}

void 
CgProgramTest::readbackFragmentShader() {
   using namespace shallows::utils;
   using namespace shallows;
   unsigned char pixels[4*16];
   for (int i=0; i<64; i++)
   {
      pixels[i]=2*i;
   }
   shared_ptr<Texture2D> tex = createUnsignedByteTexture2D
      (4, 4, 4, pixels);
   shared_ptr<Texture2D> texRect = createUnsignedByteTextureRectangle
      (4, 4, 4, pixels);

   shallows::CgProgram P;
   try { P.readFile( "shaders/frag_only.cgshader" ); }
   catch (shallows::compile_error &e)
   {
      std::cerr << e.log();
      throw;
   }
    
   CPPUNIT_ASSERT_NO_THROW( P.setInputTexture("texture", tex));
   CPPUNIT_ASSERT_NO_THROW( P.setInputTexture("textureRect", texRect));

   shared_ptr<OffScreenBuffer > fb;
   fb.reset(new OffScreenBuffer(*tex));
   shared_ptr<RenderTexture2D> renderTarget = 
      fb->createRenderTexture2D();

   CPPUNIT_ASSERT_NO_THROW( P.setFrameBuffer(fb));
   P.useNormalizedTexCoords();
    
   CPPUNIT_ASSERT_NO_THROW( P.setOutputTarget(0, renderTarget));
   CPPUNIT_ASSERT_NO_THROW( P.activate());
   glClearColor(0.5, 0.5, 0.5, 0.5);
   glClear(GL_COLOR_BUFFER_BIT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  //  gluOrtho2D(-0.5, 0.5, -0.5, 0.5);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  //  glTranslatef(-0.5f, -0.5f, 0.0f);


   CPPUNIT_ASSERT_NO_THROW( P.run());
   std::vector<unsigned char> read_pixels;
   CPPUNIT_ASSERT_NO_THROW(readUnsignedCharBuffer(*fb, *renderTarget, 
						  4, 4, 4, read_pixels));
   for (int i=0; i<16; i++)
   {
      unsigned char tmp=min(255, 2*pixels[i]);
      CPPUNIT_ASSERT_EQUAL(tmp, read_pixels[i] );
   }   

}

}; // namespace test

#endif // defined(HAVE_CG) && defined(HAVE_CGGL)
