/**********************************************************************
 * 
 *  File: FrameBufferTest.cpp
 * 
 *  Created:
 *
 *  Version: $Id: FrameBufferTest.cpp,v 1.20 2005/11/25 11:18:16 jonhjelm Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#include "FrameBufferTest.hpp"
#include "shallows/OffScreenBuffer.hpp"
#include "shallows/OnScreenBuffer.hpp"
#include "shallows/OnScreenRenderTarget.hpp"
#include "shallows/shallowsexcept.hpp"
#include "shallows/internal/GLIncludes.hpp"
#include "shallows/utils/textureCreation.hpp"
#include "shallows/utils/readBuffer.hpp"

#include "shallows/Texture2D.hpp"
#include "shallowstesting.hpp"
#include "shallows/RenderTexture2D.hpp"
#include "shallows/RenderBuffer.hpp"
#include <string>
#include <cassert> 
#include <iostream>
#include <stdexcept>

using namespace std;
using namespace shallows::internal;
using boost::shared_ptr;
namespace test
{

CPPUNIT_TEST_SUITE_REGISTRATION( FrameBufferTest );

FrameBufferTest::FrameBufferTest()
{
}

FrameBufferTest::~FrameBufferTest()
{
}


void
FrameBufferTest::setUp() {
	initGlut();
}

void
FrameBufferTest::construction() {
   shallows::Texture2D tex(GL_RGBA_FLOAT32_ATI,
			   4, 4, 
			   GL_RGBA,
			   GL_FLOAT, 
			   (const float *)0);
   CPPUNIT_ASSERT_NO_THROW( shallows::OffScreenBuffer P(tex) );	
}

void FrameBufferTest::activate() {
   shallows::Texture2D tex(GL_RGBA_FLOAT32_ATI,
			   4, 4, 
			   GL_RGBA,
			   GL_FLOAT, 
			   (const float *)0);
   shallows::OffScreenBuffer fb(tex);
   int curr_bound;
   glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); // constructor binds...
   glGetIntegerv(GL_FRAMEBUFFER_BINDING_EXT, &curr_bound);
   CPPUNIT_ASSERT_EQUAL( curr_bound, 0 );
   CPPUNIT_ASSERT_NO_THROW(fb.activate() );	
   glGetIntegerv(GL_FRAMEBUFFER_BINDING_EXT, &curr_bound);
   CPPUNIT_ASSERT_ASSERTION_FAIL( CPPUNIT_ASSERT( curr_bound == 0 ) );
}
void FrameBufferTest::deactivate() {
   CPPUNIT_ASSERT_NO_THROW( shallows::OnScreenBuffer::setActive());
   int curr_bound;
   glGetIntegerv(GL_FRAMEBUFFER_BINDING_EXT, &curr_bound);
   CPPUNIT_ASSERT_EQUAL( curr_bound, 0 );
}

void FrameBufferTest::attach2DTexture() {
   boost::shared_ptr<shallows::Texture2D> tex1;
   tex1 = shallows::utils::createFloatTexture2D(8, 8, 4, NULL);
   boost::shared_ptr<shallows::Texture2D> tex2;
   tex2 = shallows::utils::createFloatTexture2D(8, 8, 4, NULL);
   boost::shared_ptr<shallows::Texture2D> tex3;
   tex3 = shallows::utils::createFloatTexture2D(16, 16, 4, NULL);
   shallows::OffScreenBuffer P(*tex1);
   shared_ptr<shallows::RenderTexture2D> r2D1
      (new shallows::RenderTexture2D(tex1));
   shared_ptr<shallows::RenderTexture2D> r2D2
      (new shallows::RenderTexture2D(tex2));
   shared_ptr<shallows::RenderTexture2D> r2D3
      (new shallows::RenderTexture2D(tex3));
   CPPUNIT_ASSERT_NO_THROW(P.attachRenderTarget(r2D1, 0));
   CPPUNIT_ASSERT_NO_THROW(P.attachRenderTarget(r2D2, 1));
   CPPUNIT_ASSERT_THROW(P.attachRenderTarget(r2D3, 2), std::invalid_argument);
   //   CPPUNIT_ASSERT_NO_THROW( shallows::FrameBuffer::activateWindowFrameBuffer());
}

void FrameBufferTest::MRT() {
   shared_ptr<shallows::OffScreenBuffer> 
      fb(new shallows::OffScreenBuffer(4, 4));
   shared_ptr<shallows::RenderTexture2D> rt[2];
   rt[0]=fb->createRenderTexture2D();
   rt[1]=fb->createRenderTexture2D();
   CPPUNIT_ASSERT_NO_THROW(fb->attachRenderTarget(rt[0], 0));
   CPPUNIT_ASSERT_NO_THROW(fb->attachRenderTarget(rt[1], 1));
   //   CPPUNIT_ASSERT_NO_THROW(fb->drawBuffers(2));
   //   CPPUNIT_ASSERT_NO_THROW( shallows::FrameBuffer::activateWindowFrameBuffer());
}

void FrameBufferTest::validateCompleteness() {
   {
      boost::shared_ptr<shallows::Texture2D> tex1;
      tex1 = shallows::utils::createFloatTexture2D(256, 256, 4, NULL);
      shallows::OffScreenBuffer P(*tex1);
      shared_ptr<shallows::RenderTexture2D> r2D1
	 (new shallows::RenderTexture2D(tex1));
      CPPUNIT_ASSERT_NO_THROW(P.attachRenderTarget(r2D1, 0));
      CPPUNIT_ASSERT_NO_THROW(P.validateCompleteness());
   }
   
   {
      boost::shared_ptr<shallows::OffScreenBuffer> fb
	 (new shallows::OffScreenBuffer(128, 128));
      boost::shared_ptr<shallows::RenderTexture2D> rt;
      rt=fb->createRenderTexture2D();
      CPPUNIT_ASSERT_NO_THROW(fb->attachRenderTarget(rt, 0));
      CPPUNIT_ASSERT_NO_THROW(fb->validateCompleteness());
   }
}


void FrameBufferTest::depthOnly() {
   const int DIM=64;
   std::vector<float> read_depths(DIM*DIM);
   shallows::OffScreenBuffer B(DIM, DIM);
   boost::shared_ptr<shallows::RenderBuffer>  db=B.createDepthBuffer();
   B.activate();
   B.attachDepthBuffer(*db);
   glDrawBuffer(GL_NONE);
   glReadBuffer(GL_NONE);
   CPPUNIT_ASSERT_NO_THROW(B.validateCompleteness());
   //   B.setupViewport();
   glClearDepth(0.5f);
   glClear(GL_DEPTH_BUFFER_BIT);
   glReadPixels(0, 0, DIM, DIM, GL_DEPTH_COMPONENT, GL_FLOAT, 
		&read_depths[0]);
   for (int i=0; i<DIM*DIM; i++)
   {
     CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, static_cast<double> (read_depths[i]),
				  1e-3);        
   }
}

void FrameBufferTest::depthAndColor() {
   const int DIM=64;
   std::vector<float> read_depths(DIM*DIM);
   shallows::OffScreenBuffer B(DIM, DIM);
   boost::shared_ptr<shallows::RenderBuffer>  db=B.createDepthBuffer();
   B.activate();
   boost::shared_ptr<shallows::RenderBuffer>  rb=B.createRenderBuffer();
   B.attachRenderTarget(rb);
   B.attachDepthBuffer(*db);
   CPPUNIT_ASSERT_NO_THROW(B.validateCompleteness());
   //   B.setupViewport();
   glClearDepth(0.5f);
   glClear(GL_DEPTH_BUFFER_BIT);


   glReadPixels(0, 0, DIM, DIM, GL_DEPTH_COMPONENT, GL_FLOAT, 
		&read_depths[0]);
   for (int i=0; i<DIM*DIM; i++)
   {
     CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, static_cast<double> (read_depths[i]),
				  1e-3);        
   }
}

} // namespace test

