/**********************************************************************
 * 
 *  File: 
 * 
 *  Created:
 *
 *  Version: $Id: GLProgramTest.cpp,v 1.21 2005/11/25 11:18:16 jonhjelm Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#include "GLProgramTest.hpp"
#include "shallows/internal/GLIncludes.hpp"
#include "shallows/GLProgram.hpp"
#include "shallows/shallowsexcept.hpp"
#include "shallows/OffScreenBuffer.hpp"
#include "shallows/utils/textureCreation.hpp"
#include "shallows/Texture2D.hpp"
#include "shallows/RenderTexture2D.hpp"
#include "shallows/OnScreenBuffer.hpp"
#include "shallows/OnScreenRenderTarget.hpp"
#include "shallows/utils/readBuffer.hpp"
#include "shallows/utils/textureCreation.hpp"
#include "shallowstesting.hpp"
#include <string>
#include <cassert> 
#include <iostream>
#include <stdexcept>

#ifdef _WIN32
#include <windows.h>
#endif
#include <GL/gl.h>

using namespace std;
using boost::shared_ptr;
using shallows::Texture2D;
using shallows::GLProgram;
using shallows::FrameBuffer;
using shallows::RenderTexture2D;
using shallows::OffScreenBuffer;
using shallows::utils::readUnsignedCharBuffer;
using shallows::utils::readFloatBuffer;
using shallows::utils::createUnsignedByteTexture2D;
using shallows::utils::createFloatTexture2D;
namespace test
{

CPPUNIT_TEST_SUITE_REGISTRATION( GLProgramTest );

GLProgramTest::GLProgramTest()
{
}

GLProgramTest::~GLProgramTest()
{
}


void
GLProgramTest::setUp() {
	initGlut();
}

void
GLProgramTest::construction() {
	CPPUNIT_ASSERT_NO_THROW( shallows::GLProgram P );	
}

void
GLProgramTest::read() {
	// Check that we correctly read empty shader strings.
	string s = "[Vertex shader]\nvoid main() { }";
	string vs = "\nvoid main() { }";
	CPPUNIT_ASSERT( shallows::Program::getFragmentShaderSource( s ).empty() );
	string vss = shallows::Program::getVertexShaderSource( s );
	CPPUNIT_ASSERT_EQUAL( vs, vss );
	
	s = "[Fragment shader]\nvoid main() {}";
	string fs = "\nvoid main() {}";
	CPPUNIT_ASSERT( shallows::Program::getVertexShaderSource( s ).empty() );
	string fss = shallows::Program::getFragmentShaderSource( s );
	CPPUNIT_ASSERT_EQUAL( fss, fs );
	
	shallows::GLProgram P;
	
	CPPUNIT_ASSERT_THROW( P.readFile( "shaders/bogusfilename" ), shallows::file_not_found );
	CPPUNIT_ASSERT_NO_THROW( P.readFile( "shaders/showtexRECT.shader" ) );
	CPPUNIT_ASSERT( P.uniforms.size() == 1 );
	typedef shallows::Program::NameTypeMap NameTypeMap;
	CPPUNIT_ASSERT( P.uniforms.find( "foo" ) == P.uniforms.end() );
	CPPUNIT_ASSERT( P.uniforms.find( "texture" )->first == "texture" );
	CPPUNIT_ASSERT( P.uniforms.find( "texture" )->second == "sampler2DRect" );
	
	// Check that it is okay to reread some source.
	CPPUNIT_ASSERT_NO_THROW( P.readFile( "shaders/attribshader.glshader" ) );
	CPPUNIT_ASSERT( P.uniforms.find( "texture" ) == P.uniforms.end() );
	CPPUNIT_ASSERT( P.uniforms.find( "color" )->first == "color" );
	CPPUNIT_ASSERT( P.uniforms.find( "color" )->second == "sampler2D" );
}

void
GLProgramTest::setParam() {
   {
      shallows::GLProgram P;
      P.readFile( "shaders/showtexRECT.shader" );
      CPPUNIT_ASSERT_THROW( P.setParam1f( "foo", 1.0f ), std::invalid_argument );
      CPPUNIT_ASSERT_THROW( P.setParam1f( "texture", 1.0f ), std::invalid_argument );
   }
   {
      shallows::GLProgram Q;
      Q.readFile( "shaders/attribshader.glshader" );
      CPPUNIT_ASSERT_THROW( Q.setParam1f( "texture", 1.0f ), std::invalid_argument );
      CPPUNIT_ASSERT_THROW( Q.setParam1f( "color", 1.0f ), std::invalid_argument );
      CPPUNIT_ASSERT_THROW( Q.setAttrib4f( "foo", 0.0f, 0.0f, 0.0f, 0.0f ), std::invalid_argument );
      CPPUNIT_ASSERT_NO_THROW( Q.setAttrib4f( "vertexdata", 0.0f, 0.0f, 0.0f, 0.0f ) );
   }
   {
      shallows::GLProgram P;
      P.readFile( "shaders/heat.shader" );
      CPPUNIT_ASSERT_NO_THROW( P.setParam1f( "h", 1.0f ));
   }
}

void 
GLProgramTest::setInputTexture()
{
   shallows::GLProgram prog2D;
   shallows::GLProgram prog2DRect;
   boost::shared_ptr<shallows::Texture2D> tex2D;
   tex2D = shallows::utils::createFloatTexture2D(8, 8, 4, NULL);
   prog2D.readFile( "shaders/showtex2D.shader" );
   prog2DRect.readFile( "shaders/showtexRECT.shader" );
   CPPUNIT_ASSERT_NO_THROW( prog2D.setInputTexture("texture", tex2D));
   CPPUNIT_ASSERT_THROW( prog2DRect.setInputTexture("texture", tex2D),
			 std::invalid_argument);
}

void 
GLProgramTest::runSimpleProg()
{
   shallows::GLProgram prog2D;
   boost::shared_ptr<shallows::Texture2D> tex2D;
   tex2D = shallows::utils::createFloatTexture2D(8, 8, 4, NULL);
   prog2D.readFile( "shaders/showtex2D.shader" );
   CPPUNIT_ASSERT_NO_THROW( prog2D.setInputTexture("texture", tex2D));

   boost::shared_ptr<shallows::Texture2D> renderTargetTex;
   renderTargetTex = shallows::utils::createFloatTexture2D(128, 128, 4, NULL);
   boost::shared_ptr<shallows::FrameBuffer > fb;
   fb.reset(new shallows::OffScreenBuffer(*renderTargetTex));
   boost::shared_ptr<shallows::RenderTexture2D> renderTarget
      (new shallows::RenderTexture2D(renderTargetTex));
   CPPUNIT_ASSERT_NO_THROW( prog2D.setFrameBuffer(fb));

   CPPUNIT_ASSERT_NO_THROW( prog2D.setOutputTarget(0, renderTarget));
   glDisable(GL_DEPTH_TEST);
   CPPUNIT_ASSERT_NO_THROW( prog2D.run());
}

void 
GLProgramTest::runSimpleProgOnScreenBuffer()
{
   shallows::GLProgram prog2D;
   boost::shared_ptr<shallows::Texture2D> tex2D;
   tex2D = shallows::utils::createFloatTexture2D(8, 8, 4, NULL);
   prog2D.readFile( "shaders/showtex2D.shader" );
   CPPUNIT_ASSERT_NO_THROW( prog2D.setInputTexture("texture", tex2D));

   boost::shared_ptr<shallows::OnScreenBuffer> fb;
   fb.reset(new shallows::OnScreenBuffer);
   boost::shared_ptr<shallows::OnScreenRenderTarget> rt;
   rt.reset(new shallows::OnScreenRenderTarget(GL_BACK_LEFT));
   
   CPPUNIT_ASSERT_NO_THROW( prog2D.setFrameBuffer(fb));

   CPPUNIT_ASSERT_NO_THROW( prog2D.setOutputTarget(0, rt));
   CPPUNIT_ASSERT_NO_THROW( prog2D.run());
}

void 
GLProgramTest::runSimpleProgReadback()
{
   unsigned char pixels[4*16];
   for (int i=0; i<64; i++)
   {
      pixels[i]=2*i;
   }
   shared_ptr<Texture2D> init_tex = createUnsignedByteTexture2D
      (4, 4, 4, pixels);
      
   GLProgram prog2D;
   prog2D.readFile( "shaders/showtex2D.shader" );
   CPPUNIT_ASSERT_NO_THROW( prog2D.setInputTexture("texture", init_tex));
   prog2D.useNormalizedTexCoords();

   shared_ptr<Texture2D> renderTargetTex
      (new Texture2D(GL_RGBA8, 4, 4,
		     GL_RGBA, GL_UNSIGNED_BYTE,
		     NULL));
   renderTargetTex->bind();
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

   shared_ptr<FrameBuffer > fb;
   fb.reset(new OffScreenBuffer(*renderTargetTex));
   shared_ptr<RenderTexture2D> renderTarget
      (new RenderTexture2D(renderTargetTex));
   CPPUNIT_ASSERT_NO_THROW( prog2D.setFrameBuffer(fb));

   CPPUNIT_ASSERT_NO_THROW( prog2D.setOutputTarget(0, renderTarget));
   glDisable(GL_DEPTH_TEST);
   CPPUNIT_ASSERT_NO_THROW( prog2D.run());
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
   glDisable(GL_DEPTH_TEST);
   glClearColor(0.5, 0.5, 0.5, 0.5);
   glClear(GL_COLOR_BUFFER_BIT);
   CPPUNIT_ASSERT_NO_THROW( prog2D.run());

   std::vector<unsigned char> read_pixels;
   CPPUNIT_ASSERT_NO_THROW(readUnsignedCharBuffer(*fb, renderTarget, 
						  4, 4, 4, read_pixels));
   for (int i=0; i<64; i++)
   {
      CPPUNIT_ASSERT_EQUAL( pixels[i], read_pixels[i] );
   }
   CPPUNIT_ASSERT_NO_THROW(readUnsignedCharBuffer(*fb, renderTarget, 
			  3, 4, 4, read_pixels));
   for (int i=0; i<16; i++)
   {
      CPPUNIT_ASSERT_EQUAL( pixels[4*i], read_pixels[3*i] );
      CPPUNIT_ASSERT_EQUAL( pixels[4*i+1], read_pixels[3*i+1] );
      CPPUNIT_ASSERT_EQUAL( pixels[4*i+2], read_pixels[3*i+2] );
   }
   
   renderTargetTex = shallows::utils::createUnsignedByteTexture2D
      (4, 4, 3, NULL);
   fb.reset(new OffScreenBuffer(*renderTargetTex));
   renderTarget.reset(new RenderTexture2D(renderTargetTex));
   CPPUNIT_ASSERT_NO_THROW( prog2D.setFrameBuffer(fb));
   CPPUNIT_ASSERT_NO_THROW( prog2D.setOutputTarget(0, renderTarget));
   CPPUNIT_ASSERT_NO_THROW( prog2D.run());
   CPPUNIT_ASSERT_NO_THROW(readUnsignedCharBuffer(*fb, renderTarget, 
			  3, 4, 4, read_pixels));
   for (int i=0; i<16; i++)
   {
      CPPUNIT_ASSERT_EQUAL( pixels[4*i], read_pixels[3*i] );
      CPPUNIT_ASSERT_EQUAL( pixels[4*i+1], read_pixels[3*i+1] );
      CPPUNIT_ASSERT_EQUAL( pixels[4*i+2], read_pixels[3*i+2] );
   }

}

void 
GLProgramTest::activate()
{
   float pixels[4*16];
   int width, height;
   for (int i=0; i<64; i++)
   {
      pixels[i]=2*i;
   }
   shared_ptr<Texture2D> init_tex = createFloatTexture2D
      (4, 4, 4, pixels);
      
   GLProgram prog2D;
   prog2D.readFile( "shaders/showtex2D.shader" );
   CPPUNIT_ASSERT_NO_THROW( prog2D.setInputTexture("texture", init_tex));

   shared_ptr<OffScreenBuffer > fb;
   fb.reset(new OffScreenBuffer(*init_tex));
   shared_ptr<RenderTexture2D> renderTarget=fb->createRenderTexture2D();

   CPPUNIT_ASSERT_NO_THROW( prog2D.setFrameBuffer(fb));

   CPPUNIT_ASSERT_NO_THROW( prog2D.setOutputTarget(0, renderTarget));
   CPPUNIT_ASSERT_NO_THROW( prog2D.activate());
   glPushAttrib(GL_VIEWPORT_BIT);
   fb->getSize(width, height);
   glViewport(0, 0, width, height);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(-0.5, 0.5, -0.5, 0.5);
  glViewport(0,0,4,4);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(-0.5f, -0.5f, 0.0f);
   glDisable(GL_DEPTH_TEST);

   glBegin(GL_QUADS);
   glTexCoord2f(0.0, 0.0);
   glVertex3f(0.0f, 0.0f, 0.5f);

   glTexCoord2f(1.0, 0.0);
   glVertex3f(1.0f, 0.0f, 0.5f);

   glTexCoord2f(1.0, 1.0);
   glVertex3f(1.0f, 1.0f, 0.5f);

   glTexCoord2f(0.0, 1.0);
   glVertex3f(0.0f, 1.0f, 0.5f);
   glEnd();

   std::vector<float> read_pixels;
   CPPUNIT_ASSERT_NO_THROW(readFloatBuffer(*fb, renderTarget, 
					   4, 4, 4, read_pixels));
   for (int i=0; i<64; i++)
   {
      CPPUNIT_ASSERT_EQUAL( pixels[i], read_pixels[i] );
   }
   glPopAttrib();
}
void
GLProgramTest::setInputeTextureToTexunit()
{
   using namespace shallows;
   using namespace shallows::internal;
   using namespace shallows::utils;

   shared_ptr<Texture2D> tex[4];
   for (int i=0; i<4; i++)
      tex[i] = createFloatTexture2D(4, 4, 4, 0);
   GLProgram prog2D;
   prog2D.readFile( "shaders/combine_four_textures.shader" );
   // set it up incorrect
   prog2D.setInputTexture("tex1", tex[2], 1);
   prog2D.setInputTexture("tex3", tex[2]);

   // set it like it should be
   prog2D.setInputTexture("tex2", tex[2], 2);
   prog2D.setInputTexture("tex0", tex[0], 0);
   prog2D.setInputTexture("tex1", tex[1], 1);
   prog2D.setInputTexture("tex3", tex[3], 3);
   
   boost::shared_ptr<OnScreenBuffer> fb(new OnScreenBuffer);
   boost::shared_ptr<OnScreenRenderTarget> rt
      (new OnScreenRenderTarget(GL_BACK_LEFT));
   
   CPPUNIT_ASSERT_NO_THROW( prog2D.setFrameBuffer(fb));

   CPPUNIT_ASSERT_NO_THROW( prog2D.setOutputTarget(0, rt));
   CPPUNIT_ASSERT_NO_THROW( prog2D.activate());
   
   for(int i=0; i<4; i++)
   {    
      int curr_val;
      glActiveTextureARB((GLenum) (GL_TEXTURE0_ARB + i));
      glGetIntegerv(GL_TEXTURE_2D_BINDING_EXT, &curr_val);
      CPPUNIT_ASSERT_EQUAL(tex[i]->getTextureID(), 
			   static_cast<unsigned int>(curr_val));
   }


}

}; // namespace test
