/**********************************************************************
 * 
 *  File: FrameBufferTest.cpp
 * 
 *  Created:
 *
 *  Version: $Id: readBufferTest.cpp,v 1.5 2005/11/25 11:18:16 jonhjelm Exp $
 *
 *  Authors: Trond R. Hagen <trr@sintef.no>, 
 *	     Jon Mikkelsen Hjelmervik <jami@sintef.no>,
 *	     Johan S. Seland <johans@ifi.uio.no>
 *
 *  This file is part of the Shallows library.
 *  Copyright (C) 2005 by SINTEF.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 * 
 *  For using Shallows with software that can not be combined with the
 *  GNU GPL, please contact SINTEF for aquiring a commercial license
 *  and support.
 *
 *  SINTEF, Pb 124 Blindern, N-0314 Oslo, Norway
 *  http://www.sintef.no
 *********************************************************************/

#include "readBufferTest.hpp"
#include "shallows/shallows.hpp"
#include "shallows/OffScreenBuffer.hpp"
#include "shallows/OnScreenBuffer.hpp"
#include "shallows/OnScreenRenderTarget.hpp"
#include "shallows/shallowsexcept.hpp"
#include "shallows/internal/GLIncludes.hpp"
#include "shallows/utils/textureCreation.hpp"
#include "shallows/utils/readBuffer.hpp"

#include "shallows/Texture2D.hpp"
#include "shallowstesting.hpp"
#include "shallows/RenderTexture2D.hpp"
#include <string>
#include <cassert> 
#include <iostream>
#include <stdexcept>

using namespace std;
using boost::shared_ptr;
using namespace shallows;
using namespace shallows::utils;

namespace test
{

CPPUNIT_TEST_SUITE_REGISTRATION( readBufferTest );

readBufferTest::readBufferTest()
{
}

readBufferTest::~readBufferTest()
{
}


void
readBufferTest::setUp() {
	initGlut();
}

void
readBufferTest::readDefaultFrameBuffer ()
{
   glGetError();
   std::vector<float> float_pixels;
   std::vector<unsigned char> uchar_pixels;
   OffScreenBuffer fb(2, 2);
   shared_ptr<OffScreenRenderTarget> rt = fb.createRenderTexture2D();
   fb.attachRenderTarget(rt, 0);
   glPushAttrib(GL_VIEWPORT_BIT);
   //   fb.setupViewport();

   {
      float cols[]={1.0, 1.0/2.0, 1.0/4.0, 1.0/8.0};
      unsigned char  ucols[]={255, 128, 64, 32};
      glClearColor(cols[0], cols[1], cols[2], cols[3]);
      glClear(GL_COLOR_BUFFER_BIT);
      CPPUNIT_ASSERT_NO_THROW(readFloatBuffer(fb, rt, 4, 2, 2, float_pixels));
      for (unsigned int i=0; i<4; i++)
      {
	 for (unsigned int j=0; j<4; j++)
	 {
	    CPPUNIT_ASSERT_EQUAL(cols[j], float_pixels[4*i+j]);
	 }
      }
      CPPUNIT_ASSERT_NO_THROW(readUnsignedCharBuffer
			      (fb, rt, 4, 2, 2, uchar_pixels));
      for (unsigned int i=0; i<4; i++)
      {
	 for (unsigned int j=0; j<4; j++)
	 {
	    CPPUNIT_ASSERT_EQUAL(ucols[j], uchar_pixels[4*i+j]);
	 }
      }
      
   }

   glPopAttrib();
}

void clearAndRead (unsigned int components, 
				      bool rectangle)
{
   unsigned int height = 4;
   unsigned int width = 4;
   std::vector<float> float_pixels;
   std::vector<unsigned char> uchar_pixels;
   shared_ptr<OffScreenBuffer> fb;
   shared_ptr<OffScreenRenderTarget> rt;
   shared_ptr<Texture2D> tex;
   float cols[]={1.0, 1.0/2.0, 1.0/4.0, 1.0/8.0};
   unsigned char ucols[]={255, 128, 64, 32};

   
   float use_cols[4];
   unsigned char use_ucols[4];
   
   if (components>=3)
   {
      for (int i=0; i<components; i++)
	 use_cols[i]=cols[i];
   }
   else if (components==2)
   {
      use_cols[0]=cols[0];
      for (int i=1; i<4; i++)
	 use_cols[i]=cols[1];
   } 
   else if (components==1)
   {
      use_cols[0]=cols[0];
      for (int i=0; i<4; i++)
	 use_cols[i]=cols[0];
   } 

   if (rectangle)
      tex = createFloatTextureRectangle(width, height,  components, 0);
   else
      tex = createFloatTexture2D(width, height,  components, 0);
   fb.reset(new OffScreenBuffer(*tex));
   if (rectangle)
      rt = fb->createRenderTextureRectangle();
   else
      rt = fb->createRenderTexture2D();
   fb->attachRenderTarget(rt, 0);
   fb->validateCompleteness();
   glPushAttrib(GL_VIEWPORT_BIT);
   //   fb->setupViewport();

   {
      glClearColor(use_cols[0], use_cols[1], use_cols[2], use_cols[3]);
      glClear(GL_COLOR_BUFFER_BIT);
      CPPUNIT_ASSERT_NO_THROW(readFloatBuffer
			      (*fb, rt, components, width, height, 
			       float_pixels));
      for (unsigned int i=0; i<width*height; i++)
      {
	 for (unsigned int j=0; j<components; j++)
	 {
	    CPPUNIT_ASSERT_EQUAL(cols[j], float_pixels[components*i+j]);
	 }
      }
      CPPUNIT_ASSERT_NO_THROW(readUnsignedCharBuffer
      			      (*fb, rt, components, width, height, 
			       uchar_pixels));
        for (unsigned int i=0; i<width*height; i++)
        {
  	 for (unsigned int j=0; j<components; j++)
  	 {
	    CPPUNIT_ASSERT_EQUAL(ucols[j], uchar_pixels[components*i+j]);
  	 }
        }
   }

   glPopAttrib();
}

void 
readBufferTest::readOffScreenFrameBuffer ()
{
   unsigned int components = 2;

   for (components = 3; components<=4; components++)
   {
      clearAndRead (components, false);
   }

   for (components = 1; components<=4; components++)
   {
      clearAndRead (components, true);
   }
}

void 
readBufferTest::readOnScreenFrameBuffer ()
{
   boost::shared_ptr<OnScreenBuffer> fb(new OnScreenBuffer);
   boost::shared_ptr<OnScreenRenderTarget> rt
      (new OnScreenRenderTarget(GL_BACK_LEFT));
   std::vector<unsigned char> read_pixels;
   float cols[]={1.0, 1.0/2.0, 1.0/4.0, 1.0/8.0};
   unsigned char ucols[]={255, 128, 64, 32};
   glClearColor(cols[0], cols[1], cols[2], cols[3]);
   glClear(GL_COLOR_BUFFER_BIT);
   CPPUNIT_ASSERT_NO_THROW
      (shallows::utils::readUnsignedCharBuffer(*fb, rt, 3, 4, 4, 
					       read_pixels));
   for (unsigned int i=0; i<4*4; i++)
   {
      for (unsigned int j=0; j<3; j++)
      {
	 CPPUNIT_ASSERT_EQUAL(ucols[j], read_pixels[3*i+j]);
      }
   }
}

void 
readBufferTest::readRenderBuffer ()
{
   glGetError();
   CHECK_GL_DEBUG;
   boost::shared_ptr<OffScreenBuffer> fb;
// Would like a smaller framebuffer, but it seg.faults...
   const int width=128;
   const int height=128;
   fb.reset(new OffScreenBuffer(GL_RGBA8, width, height)); 
   boost::shared_ptr<RenderBuffer> rt;
   rt = fb->createRenderBuffer();
   std::vector<unsigned char> read_pixels;
   float cols[]={1.0, 1.0/2.0, 1.0/4.0, 1.0/8.0};
   unsigned char ucols[]={255, 128, 64, 32};

   fb->attachRenderTarget(rt);
   fb->validateCompleteness();
   glPushAttrib(GL_VIEWPORT_BIT);
   {
      //      fb->setupViewport();

      glClearColor(cols[0], cols[1], cols[2], cols[3]);
      glClear(GL_COLOR_BUFFER_BIT);
      CPPUNIT_ASSERT_NO_THROW
	 (shallows::utils::readUnsignedCharBuffer(*fb, rt, 4, 
						  width, height, 
						  read_pixels));
      for (unsigned int i=0; i<width*height; i++)
      {
	 for (unsigned int j=0; j<4; j++)
	 {
	   CPPUNIT_ASSERT_EQUAL(ucols[j], read_pixels[4*i+j]);
	 }
      }
   }
   glPopAttrib();
}

} // namespace test

